
import collections
import collections.abc
import copy
import inspect
import operator
import uuid
from .logger import logger

ELEMENT_TYPE_LIST = [int, float, str]

try:
    import numpy
except Exception:
    pass
else:
    ELEMENT_TYPE_LIST = ELEMENT_TYPE_LIST+[numpy.ndarray]


class LazyProxyHandler:
    @classmethod
    def put(cls, obj, path, value, *args, **kwargs):
        for p in path[:-1]:
            if type(p) is str and hasattr(obj, p):
                obj = getattr(obj, p)
            elif p in obj:
                obj = obj[p]
            else:
                obj[p] = {}
                obj = obj[p]
        if isinstance(path[-1], str) and hasattr(obj, path[-1]):
            setattr(obj, path[-1], value)
        else:
            obj[path[-1]] = value

        # if len(path) > 0:
        #     raise path

        return None

    @classmethod
    def get(cls, obj, path, *args, **kwargs):
        for idx, p in enumerate(path):
            if type(p) is str and hasattr(obj, p):
                obj = getattr(obj, p)
            else:
                if hasattr(obj.__class__, "__getitem__"):
                    obj = obj[p]
                elif isinstance(p, int) and p == 0:
                    pass
                else:
                    raise KeyError(path[:idx]+[p])

                # try:
                #     obj = obj[p]
                # except IndexError:
                #     raise KeyError(path)
                # except TypeError:
                #     raise KeyError(path)
        return obj

    @classmethod
    def get_value(cls, obj, path, *args, **kwargs):
        return cls.get(obj, path, *args, **kwargs)

    @classmethod
    def delete(cls, obj, path, *args, **kwargs):
        if len(path) > 1:
            obj = cls.get(obj,  path[:-1], *args, **kwargs)

        if hasattr(obj, path[-1]):
            delattr(obj, path[-1])
        else:
            del obj[path[-1]]

    @classmethod
    def push_back(cls, obj,  path, value, *args, **kwargs):
        if len(path) > 0:
            obj = cls.get_value(obj, path[:-1]).setdefault(path[-1], [])
        obj.append(value)
        return path+[len(obj)-1]

    @classmethod
    def count(cls, obj,  path, *args, **kwargs):
        obj = cls.get(obj, path, *args, **kwargs)
        return len(obj)

    @classmethod
    def resize(cls, obj,  path, *args, **kwargs):
        obj = cls.get(obj, path, *args, **kwargs)
        return len(obj)

    @classmethod
    def contains(cls, obj,  path, v, *args, **kwargs):
        obj = cls.get(obj, path, *args, **kwargs)
        return v in obj

    @classmethod
    def call(cls, obj, path, *args, **kwargs):
        obj = cls.get(obj, path)
        if callable(obj):
            return obj(*args, **kwargs)
        elif len(args)+len(kwargs) == 0:
            return obj
        else:
            raise TypeError(f"{obj.__class__.__name__} is not callable")

    @classmethod
    def iter(cls, obj, path, *args, **kwargs):
        for o in cls.get(obj, path, *args, **kwargs):
            if type(o) in ELEMENT_TYPE_LIST:
                yield o
            else:
                yield LazyProxy(o)


class LazyProxy:
    """ This is a lazy proxy to __getattr__ and __getitem__.
        The attribute operator (__getattr__) or index operator (__getitem__) of object is
        cached and chained as a path. Real value is only fetched when object is actually used.
        * inspired by lazy_object_proxy
        Example:
            >>> obj = {}
            >>> obj_proxy = LazyProxy(obj)
            >>> obj_proxy.e.f.d = "hello World"
            >>> self.assertEqual(obj, {"e": {"f": {"d": "hello World"}}})

        Example:
            >>> obj = {"a": {"b": {"f": 5, "c": {"d": [1, 2, 3, 4, 5]}}},
                "msg": "hello World"}

            >>> obj_proxy = LazyProxy(obj)

            >>> self.assertEqual(obj_proxy.a.b.c.d[1:5], obj["a"]["b"]["c"]["d"][1:5])
            >>> self.assertEqual(obj_proxy.a["b.c.d"][1:5],obj["a"]["b"]["c"]["d"][1:5])
            >>> self.assertEqual(obj_proxy["a.b.c.d[:]"], obj["a"]["b"]["c"]["d"][1:5])
            >>> self.assertEqual(obj_proxy.a.b.f*3, obj["a"]["b"]["f"]*3)

                equilibrium.time_slice[0].profiles_2d[0].psi[:,:] =>   PROFILES[0].PSIRZ[0,:,:]
                equilibrium.time_slice[:].profiles_2d[1].psi[:,:] =>   PROFILES[1].PSIRZ[:,:,:]

                [
                    {psi:345.7, a:1},
                    {psi:5.68,  a:4}
                ]
                =>
                {
                    psi:[345.7,5.678],
                    a  :[1,4]
                }

        Example:
            >>>   class Foo:
                __slots__ = 'a', 'b'

            >>> obj = Foo()
            >>> obj.a = [1, 2, 3, 4, 5]
            >>> obj.b = {"a": {"b": {"f": 5, "c": {"d": [1, 2, 3, 4, 5]}}},
                    "msg": "hello World"}
            >>> obj_proxy = LazyProxy(obj)

            >>> self.assertEqual(obj_proxy["a"][0]*3, obj.a[0]*3)
            >>> self.assertEqual(obj_proxy.b.a.b.c.d[1:5], obj.b["a"]["b"]["c"]["d"][1:5])

        TODO: support add/append
    """
    __slots__ = "__object__", "__path__", "__level__", "__handler__"

    DELIMITER = '.'

    @staticmethod
    def wrap(obj, handler=None, **kwargs):
        if handler is None and hasattr(obj.__class__, "__lazy_proxy__"):
            handler = obj.__class__.__lazy_proxy__

        if isinstance(handler, LazyProxyHandler) or inspect.isclass(handler):
            return handler
        elif isinstance(handler, (type(None), collections.abc.Mapping)):
            handler = collections.ChainMap(handler or {}, kwargs)
        elif inspect.isfunction(handler):
            handler = collections.ChainMap({"get": handler}, kwargs)
        else:
            raise TypeError(f"illegal ops type {type(handler)}!")

        n_cls = type(
            f"wrapped_{obj.__class__.__name__}_{uuid.uuid1().hex}",
            (LazyProxyHandler,),
            {k: classmethod(op) for k, op in handler.items()}
        )
        return n_cls()

    def __init__(self,   *args,    **kwargs):
        self.__reset__(*args, **kwargs)

    def __reset__(self, obj, prefix=[], *args, level=-1, handler=None, **kwargs):
        if prefix is None:
            prefix = []
        elif self.__class__.DELIMITER is not None and type(prefix) is str:
            prefix = prefix.split(self.__class__.DELIMITER)
        elif not isinstance(prefix, list):
            prefix = [prefix]

        if isinstance(obj, LazyProxy):
            object.__setattr__(self, "__object__",
                               object.__getattribute__(obj, "__object__"))
            object.__setattr__(self, "__path__",
                               object.__getattribute__(obj, "__path__")+prefix)
            object.__setattr__(self, "__level__",
                               object.__getattribute__(obj, "__level__")-1)
            object.__setattr__(self, "__handler__",
                               object.__getattribute__(obj, "__handler__")
                               if handler is None and len(kwargs) == 0 else LazyProxy.wrap_handler(obj, handler, **kwargs))

        else:
            object.__setattr__(self, "__object__", obj)
            object.__setattr__(self, "__path__", prefix)
            object.__setattr__(self, "__level__", level)
            object.__setattr__(self, "__handler__",  LazyProxy.wrap(obj, handler, **kwargs))

    def __fetch__(self):
        obj = object.__getattribute__(self, "__object__")
        path = object.__getattribute__(self, "__path__")
        handler = object.__getattribute__(self, "__handler__")

        if len(path) == 0:
            pass
        else:
            try:
                obj = handler.get(obj, path)
            except KeyError:
                raise KeyError(f"Unsolved path '{path}'")
            else:
                object.__setattr__(self, "__path__", [])
                object.__setattr__(self, "__object__", obj)

        return obj

    def __push_back__(self, v=None):
        obj = object.__getattribute__(self, "__object__")
        path = object.__getattribute__(self, "__path__")
        handler = object.__getattribute__(self, "__handler__")
        new_path = handler.push_back(obj, path, v)
        return LazyProxy(obj, prefix=new_path, handler=handler)

    def __pop_back__(self):
        obj = object.__getattribute__(self, "__object__")
        path = object.__getattribute__(self, "__path__")
        handler = object.__getattribute__(self, "__handler__")
        return handler.pop_back(obj, path)

    def __real_value__(self):
        obj = object.__getattribute__(self, "__object__")
        path = object.__getattribute__(self, "__path__")
        handler = object.__getattribute__(self, "__handler__")
        return handler.get_value(obj, path)

    def _value_(self):
        res = self.__real_value__()
        if isinstance(res, collections.abc.Mapping):
            return LazyProxy(res)
        else:
            return res

    def __do_get__(self, idx):
        res = LazyProxy(self, idx)
        return res.__fetch__() if self.__level__ == 0 else res

    def __do_set__(self, idx, value):
        object.__getattribute__(self, "__handler__").put(
            object.__getattribute__(self, "__object__"),
            object.__getattribute__(self, "__path__")+[idx],
            value
        )

    def __do_del__(self, idx):
        object.__getattribute__(self, "__handler__").delete(
            object.__getattribute__(self, "__path__")+[idx]
        )

    def __getitem__(self, idx):
        return self.__do_get__(idx)

    def __setitem__(self, idx, value):
        self.__do_set__(idx, value)

    def __delitem__(self, idx):
        self.__do_del__(idx)

    def __getattr__(self, idx):
        return self.__do_get__(idx)

    def __setattr__(self, idx, value):
        self.__do_set__(idx, value)

    def __delattr__(self, idx):
        self.__do_del__(idx)

    def __deepcopy__(self, memo=None):
        handler = object.__getattribute__(self, "__handler__")
        return LazyProxy(copy.deepcopy(self.__fetch__()),    handler=handler)

    def __copy__(self):
        handler = object.__getattribute__(self, "__handler__")
        return LazyProxy(copy.copy(self.__fetch__()),   handler=handler)

    # def __getslice__(self, i, j):
    #     return self.__do_get__(slice(i, j))

    # def __setslice__(self, i, j, value):
    #     self.__do_set__(slice(i, j), value)

    # def __delslice__(self, i, j):
    #     self.__do_del__(slice(i, j))

    def __iter__(self):
        return object.__getattribute__(self, "__handler__").iter(
            object.__getattribute__(self, "__object__"),
            object.__getattribute__(self, "__path__")
        )

    def __len__(self):
        return object.__getattribute__(self, "__handler__").count(
            object.__getattribute__(self, "__object__"),
            object.__getattribute__(self, "__path__")
        )

    def __resize__(self, *args, **kwargs):
        return object.__getattribute__(self, "__handler__").resize(
            object.__getattribute__(self, "__object__"),
            object.__getattribute__(self, "__path__"),
            * args, **kwargs)

    def __contains__(self, value):
        return object.__getattribute__(self, "__handler__").contains(
            object.__getattribute__(self, "__object__"),
            object.__getattribute__(self, "__path__"),
            value
        )

    def __call__(self, *args, **kwargs):

        res = self.__fetch__()
        if callable(res):
            return res(*args, **kwargs)
        else:
            return res
        # return object.__getattribute__(self, "__handler__").call(
        #     object.__getattribute__(self, "__object__"),
        #     object.__getattribute__(self, "__path__"),
        #     *args, **kwargs
        # )

    ###############################################################
    # from lazy_object_proxy

    # @property
    # def __name__(self):
    #     return self.__fetch__().__name__

    # @__name__.setter
    # def __name__(self, value):
    #     self.__fetch__().__name__ = value

    # @property
    # def __class__(self):
    #     return self.__fetch__().__class__

    # @property
    # def __annotations__(self):
    #     return self.__fetch__().__anotations__

    # def __dir__(self):
    #     return dir(self.__fetch__())

    def __str__(self):
        return str(self.__fetch__())

    def __bytes__(self):
        return bytes(self.__fetch__())

    def __reversed__(self):
        return reversed(self.__fetch__())

    def __round__(self):
        return round(self.__fetch__())

    def __hash__(self):
        return hash(self.__fetch__())

    def __nonzero__(self):
        return bool(self.__fetch__())

    def __bool__(self):
        return bool(self.__fetch__())

    def __int__(self):
        return int(self.__fetch__())

    # if PY2:
    #     def __long__(self):
    #         return long(self.__fetch__())  # noqa

    def __float__(self):
        return float(self.__fetch__())

    def __oct__(self):
        return oct(self.__fetch__())

    def __hex__(self):
        return hex(self.__fetch__())

    def __index__(self):
        return operator.index(self.__fetch__())

    def __enter__(self):
        return self.__fetch__().__enter__()

    def __exit__(self, *args, **kwargs):
        return self.__fetch__().__exit__(*args, **kwargs)

    def __reduce__(self):
        return lambda p: p, (self.__fetch__(),)

    def __reduce_ex__(self, protocol):
        return lambda p: p, (self.__fetch__(),)


__op_list__ = ['abs', 'add', 'and',
               #  'attrgetter',
               'concat',
               # 'contains', 'countOf',
               'delitem', 'eq', 'floordiv', 'ge',
               # 'getitem',
               'gt',
               'iadd', 'iand', 'iconcat', 'ifloordiv', 'ilshift', 'imatmul', 'imod', 'imul',
               'index', 'indexOf', 'inv', 'invert', 'ior', 'ipow', 'irshift',
               #    'is_', 'is_not',
               'isub',
               # 'itemgetter',
               'itruediv', 'ixor', 'le',
               'length_hint', 'lshift', 'lt', 'matmul',
               #    'methodcaller',
               'mod',
               'mul', 'ne', 'neg', 'not', 'or', 'pos', 'pow', 'rshift',
               #    'setitem',
               'sub', 'truediv', 'truth', 'xor']


for name in __op_list__:
    op = getattr(operator, f"__{name}__", None)
    setattr(LazyProxy,  f"__{name}__",
            lambda s, r, __fun__=op: __fun__(s.__fetch__(), r))
    setattr(LazyProxy,  f"__r{name}__",
            lambda s, l, *args, __fun__=op, **kwargs: __fun__(l, s.__fetch__(), *args, **kwargs))



from __future__ import annotations

import collections.abc
import functools
import typing
from copy import copy

import numpy as np
import scipy.constants
from scipy.interpolate import CubicSpline, PPoly

from ..utils.logger import logger
from ..utils.typing import (ArrayLike, ArrayType, NumericType, array_type,
                            nTupleType)
from .BBox import BBox
from .GeoObject import GeoObject
from .PointSet import PointSet


@GeoObject.register(["curve", "Curve"])
class Curve(PointSet):
    """ Curve
        曲线，一维几何体
    """

    def __init__(self, *args, uv=None, **kwargs) -> None:
        super().__init__(*args, rank=1, ** kwargs)
        self._metadata["closed"] = np.allclose(self._points[0], self._points[-1])
        self._uv = uv if uv is not None else np.linspace(0, 1.0, self._points.shape[0])

    def __copy__(self) -> Curve:
        other: Curve = super().__copy__()  # type:ignore
        other._uv = self._uv
        return other

    @functools.cached_property
    def dl(self) -> array_type:
        x, y = self.points
        a, b = self.derivative()

        # a = a[:-1]
        # b = b[:-1]
        dx = x[1:]-x[:-1]
        dy = y[1:]-y[:-1]

        m1 = (-a[:-1]*dy+b[:-1]*dx)/(a[:-1]*dx+b[:-1]*dy)

        # a = np.roll(a, 1, axis=0)
        # b = np.roll(b, 1, axis=0)

        m2 = (-a[1:]*dy+b[1:]*dx)/(a[1:]*dx+b[1:]*dy)

        return np.sqrt(dx**2+dy**2)*(1 + (2.0*m1**2+2.0*m2**2-m1*m2)/30)

    @functools.cached_property
    def measure(self) -> float: return np.sum(self.dl)

    def integral(self, func: typing.Callable) -> float:
        x, y = self.points
        val = func(x, y)

        # c_pts = self.points((self._mesh[0][1:] + self._mesh[0][:-1])*0.5)

        return np.sum(0.5*(val[:-1]+val[1:]) * self.dl)

    def coordinates(self, *uvw, **kwargs) -> ArrayType:
        if len(uvw) == 0:
            return self._points
        else:
            return self._spl(*uvw, **kwargs)

    @functools.cached_property
    def _spl(self) -> PPoly:
        return CubicSpline(self._uv, self._points, bc_type="periodic" if self.is_closed else "not-a-knot")

    @functools.cached_property
    def _derivative(self):
        return self._spl.derivative()

    def derivative(self, *args, **kwargs):
        if len(args) == 0:
            args = [self._uv]
        res = self._derivative(*args, **kwargs)
        return res[:, 0], res[:, 1]

    def enclose(self, *args) -> bool:
        if not self.is_closed:
            return False
        return super().enclose(*args)

    def remesh(self, u) -> Curve:
        other: Curve = copy(self)
        if isinstance(u, array_type):
            other._uv = u
        elif callable(u):
            other._uv = u(*self.points)
        else:
            raise TypeError(f"illegal type u={type(u)}")
        return other



import os

from spdm.utils.logger import logger
from spdm.core.Entry import open_db


os.environ["SP_DATA_MAPPING_PATH"] = "/home/salmon/workspace/fytok_data/mapping"

if __name__ == '__main__':

    # db = open_db("mdsplus[EAST]:///home/salmon/workspace/data/~t/?tree_name=efit_east")

    # entry = db.find_one(38300)

    # logger.debug(entry.get(["pf_active"]).dump())

    db = open_db("MDSplus[EAST]://202.127.204.12?tree_name=efit_east")

    entry = db.find_one(114730)

    logger.debug(entry.child("pf_active").__value__())



import collections
import collections.abc
import typing
from functools import cached_property

import numpy as np

from ..core.Function import Function
from ..geometry.BBox import BBox
from ..geometry.Box import Box
from ..geometry.Curve import Curve
from ..geometry.GeoObject import GeoObject
from ..geometry.Line import Line
from ..geometry.Point import Point
from ..numlib.interpolate import interpolate
from ..utils.logger import logger
from ..utils.typing import ArrayType, NumericType, ScalarType, array_type, numeric_type, scalar_type
from .Mesh import Mesh
from .mesh_structured import StructuredMesh

# from scipy.interpolate import (InterpolatedUnivariateSpline,
#                                RectBivariateSpline, RegularGridInterpolator,
#                                UnivariateSpline, interp1d, interp2d)


@Mesh.register("rectilinear")
class RectilinearMesh(StructuredMesh):
    """    A `rectilinear Mesh` is a tessellation by rectangles or rectangular cuboids (also known as rectangular parallelepipeds)
    that are not, in general, all congruent to each other. The cells may still be indexed by integers as above, but the
    mapping from indexes to vertex coordinates is less uniform than in a regular Mesh. An example of a rectilinear Mesh
    that is not regular appears on logarithmic scale graph paper.
    -- [https://en.wikipedia.org/wiki/Regular_Mesh]

    RectlinearMesh

    可以视为由 n=rank 条称为axis的曲线 curve 平移张成的空间。

    xyz= sum([ axis[i](uvw[i]) for i in range(rank) ])

    """

    def __init__(self, *args: ArrayType, geometry=None, periods=None, dims=None, **kwargs) -> None:
        if dims is None:
            dims = args
        elif len(args) > 0:
            raise RuntimeError(f"ignore args {args}")

        if geometry is None:
            geometry = Box([min(d) for d in dims], [max(d) for d in dims])

        for idx in range(len(dims)):
            if (
                isinstance(periods, collections.abc.Sequence)
                and periods[idx] is not None
                and not np.isclose(dims[idx][-1] - dims[idx][0], periods[idx])
            ):
                raise RuntimeError(
                    f"idx={idx} periods {periods[idx]} is not compatible with dims [{dims[idx][0]},{dims[idx][-1]}] "
                )
            if not np.all(dims[idx][1:] > dims[idx][:-1]):
                raise RuntimeError(f"dims[{idx}] is not increasing")

        super().__init__(shape=[len(d) for d in dims], geometry=geometry, **kwargs)
        self._dims = dims
        self._periods = periods
        self._aixs = [Function(self._dims[i], np.linspace(0, 1.0, self.shape[i])) for i in range(self.rank)]

    @property
    def dim1(self) -> ArrayType:
        return self._dims[0].__array__()

    @property
    def dim2(self) -> ArrayType:
        return self._dims[1].__array__()

    @property
    def dims(self) -> typing.List[ArrayType]:
        return self._dims

    @property
    def dimensions(self) -> typing.List[ArrayType]:
        return self._dims

    @property
    def rank(self) -> int:
        return len(self._dims)

    @cached_property
    def dx(self) -> ArrayType:
        return np.asarray([(d[-1] - d[0]) / len(d) for d in self._dims])

    def coordinates(self, *uvw) -> ArrayType:
        """网格点的 _空间坐标_
        @return: _数组_ 形状为 [geometry.dimension,<shape of uvw ...>]
        """
        if len(uvw) == 1 and self.rank != 1:
            uvw = uvw[0]
        return np.stack([self._dims[i](uvw[i]) for i in range(self.rank)], axis=-1)

    @cached_property
    def vertices(self) -> ArrayType:
        """网格点的 _空间坐标_"""
        if self.geometry.rank == 1:
            return (self._dims[0],)
        else:
            return np.stack(self.points, axis=-1)

    @cached_property
    def points(self) -> typing.List[ArrayType]:
        """网格点的 _空间坐标_"""
        if self.geometry.rank == 1:
            return (self._dims[0],)
        else:
            return np.meshgrid(*self._dims, indexing="ij")

    def interpolate(self, value: ArrayType, **kwargs):
        """生成插值器
        method: "linear",   "nearest", "slinear", "cubic", "quintic" and "pchip"
        """

        if not isinstance(value, np.ndarray):
            raise ValueError(f"value must be np.ndarray, but {type(value)} {value}")

        elif tuple(value.shape) != tuple(self.shape):
            raise NotImplementedError(f"{value.shape}!={self.shape}")

        if np.any(tuple(value.shape) != tuple(self.shape)):
            raise ValueError(f"{value} {self.shape}")

        return interpolate(*self._dims, value, periods=self._periods, **kwargs)




import os

from spdm.core.File import File
from spdm.flow.ModuleRepository import ModuleRepository
from spdm.flow.SpModule import SpModule
from ..util.logger import logger
import pprint

if __name__ == "__main__":

    logger.info("====== START =======")

    module = ModuleRepository(repo_name='FuYun', repo_tag='FY')

    module.load_configure("/home/salmon/workspace/fytok/external/SpDB/examples/data/FuYun/configure.yaml")

    os.environ["SP_OUTPUT_DIR"] = "/home/salmon/workspace/output"

    module.build("physics/cql3d")

    # CQL3D = module.new_class("physics/cql3d")

    # Genray = module.new_class("physics/genray", version="10.13_200117", tag="-gompi-2020a")

    # logger.debug(pprint.pformat((CQL3D._metadata)))

    # logger.debug(pprint.pformat((Genray._metadata)))

    # cfg = {
    #     "$schema": "file/namelist",
    #     "default": {
    #         "tokamak": {
    #             # {"$class": "file.geqdsk", "path": "{FY_MODULEFILE_DIR}/template/g063982.04800"},
    #             "eqdskin": "{equilibrium}"
    #         },
    #         "genr": {
    #             "partner":  {"$class": "file.netcdf", "path": "/home/salmon/workspace/data/genray/genray_profs_in.nc"},
    #             "outdat": "{WORKING_DIR}"
    #         },
    #         "ecocone": {"gzone": 1}
    #     }
    # }
    # equilibrium = File("/home/salmon/workspace/fytok/external/SpDB/examples/data/FuYun/modules/physics/genray/10.13_200117-gompi-2020a/template/g063982.04800",
    #                   format="geqdsk")

    # genray = Genray(num_of_steps=1, dt=0.001, equilibrium=equilibrium,  config=cfg)

    # logger.debug(genray.outputs.EXITCODE)

    # cql3d = CQL3D(equilibrium=genray.outputs.equilibrium, ray_trace=genray.outputs.result)

    # logger.debug(cql3d.outputs.EXITCODE)

    # out_nc = genray_out.out_nc
    # out_eq = genray_out.out_eq

    # wr = out_nc.entry.wr[1]/100.0
    # wz = out_nc.entry.wz[:]/100.0

    # plt.plot(out_eq.limrz[:, 0], out_eq.limrz[:, 1])
    # plt.plot(out_eq.bbsrz[:, 0], out_eq.bbsrz[:, 1])

    # plt.plot(wr, wz)
    # plt.contour(
    #     out_nc.eqdsk_r[:, :],
    #     out_nc.eqdsk_z[:, :],
    #     out_nc.eqdsk_psi[:, :])

    # plt.savefig("../output/demo_module.svg")

    logger.info("====== Done =======")



from __future__ import annotations

from copy import copy, deepcopy
import collections.abc
import functools
import typing
from enum import Enum
import numpy as np
import numpy.typing as np_tp

from spdm.utils.typing import array_type

from ..core.Domain import DomainBase
from ..mesh.Mesh import Mesh
from ..utils.logger import logger
from ..utils.misc import group_dict_by_prefix
from ..utils.tags import _not_found_
from ..utils.typing import ArrayType, array_type, as_array, is_array
from ..utils.numeric import float_nan, meshgrid, bitwise_and

from .Expression import Expression
from .Functor import Functor
from .Path import update_tree, Path


def guess_mesh(holder, prefix="mesh", **kwargs):
    if holder is None or holder is _not_found_:
        return None

    metadata = getattr(holder, "_metadata", {})

    mesh, *_ = group_dict_by_prefix(metadata, prefix, sep=None)

    if mesh is None:
        coordinates, *_ = group_dict_by_prefix(metadata, "coordinate", sep=None)

        if coordinates is not None:
            coordinates = {int(k): v for k, v in coordinates.items() if k.isdigit()}
            coordinates = dict(sorted(coordinates.items(), key=lambda x: x[0]))
            coordinates = [Path(c).find(holder) for c in coordinates.values()]
            if all([is_array(c) for c in coordinates]):
                mesh = {"dims": coordinates}

    elif isinstance(mesh, str) and mesh.isidentifier():
        mesh = getattr(holder, mesh, _not_found_)
    elif isinstance(mesh, str):
        mesh = Path(mesh).get(holder, _not_found_)
    elif isinstance(mesh, Enum):
        mesh = {"type": mesh.name}

    elif isinstance(mesh, collections.abc.Sequence) and all(isinstance(d, array_type) for d in mesh):
        mesh = {"dims": mesh}

    elif isinstance(mesh, collections.abc.Mapping):
        pass

    if mesh is None or mesh is _not_found_:
        return guess_mesh(getattr(holder, "_parent", None), prefix=prefix, **kwargs)
    else:
        return mesh

    # if all([isinstance(c, str) and c.startswith("../grid") for c in coordinates.values()]):
    #     o_mesh = getattr(holder, "grid", None)
    #     if isinstance(o_mesh, Mesh):
    #         # if self._mesh is not None and len(self._mesh) > 0:
    #         #     logger.warning(f"Ignore {self._mesh}")
    #         self._domain = o_mesh
    #     elif isinstance(o_mesh, collections.abc.Sequence):
    #         self._domain = update_tree_recursive(self._domain, {"dims": o_mesh})
    #     elif isinstance(o_mesh, collections.abc.Mapping):
    #         self._domain = update_tree_recursive(self._domain, o_mesh)
    #     elif o_mesh is not None:
    #         raise RuntimeError(f"holder.grid is not a Mesh, but {type(o_mesh)}")
    # else:
    #     dims = tuple([(holder.get(c) if isinstance(c, str) else c) for c in coordinates.values()])
    #     self._domain = update_tree_recursive(self._domain, {"dims": dims})


class Field(Expression):
    """Field

    Field 是 Function 在流形（manifold/Mesh）上的推广， 用于描述流形上的标量场，矢量场，张量场等。

    Field 所在的流形记为 mesh ，可以是任意维度的，可以是任意形状的，可以是任意拓扑的，可以是任意坐标系的。

    Mesh 网格描述流形的几何结构，比如网格的拓扑结构，网格的几何结构，网格的坐标系等。

    Field 与 Function的区别：
        - Function 的 mesh 是一维数组表示dimensions/axis
        - Field 的 mesh 是 Mesh，可以表示复杂流形上的场等。
    """

    Domain = Mesh

    def __init__(self, *xy, **kwargs):
        mesh, kwargs = group_dict_by_prefix(kwargs, prefixes="mesh")

        super().__init__(None, **kwargs)

        if len(xy) == 0:
            raise RuntimeError(f"illegal x,y {xy} ")

        value = xy[-1]

        dims = xy[:-1]

        if isinstance(value, (Functor, Expression)) or callable(value):
            func = value
            value = None
        else:
            func = None
            value = as_array(value)

        if mesh is not None:
            if len(dims) == 0:
                pass
            elif isinstance(mesh, dict):
                mesh["dims"] = dims
            else:
                raise RuntimeError(f"'mesh' is defined, ignore dims={dims} {mesh}")

        elif "domain" in kwargs:
            mesh = kwargs["domain"]
            if len(dims) > 0:
                raise RuntimeError(f"'mesh' is defined, ignore dims={dims}")
        else:
            mesh = dims

        self._op = func
        self._mesh = mesh
        self._cache = value
        self._ppoly = None

    def __geometry__(self, view_point="RZ", **kwargs):
        """
        plot o-point,x-point,lcfs,separatrix and contour of psi
        """

        geo = {}

        match view_point.lower():
            case "rz":
                geo["$data"] = (*self.mesh.points, self.__array__())
                geo["$styles"] = {
                    "label": self.__label__,
                    "axis_label": self.mesh.axis_label,
                    "$matplotlib": {"levels": 40, "cmap": "jet"},
                }
        return geo

    def _repr_svg_(self) -> str:
        from ..view.View import display

        return display(self.__geometry__(), output="svg")

    @property
    def mesh(self) -> Mesh:
        if isinstance(self._mesh, Mesh):
            return self._mesh

        if self._mesh is None or self._mesh is _not_found_ or len(self._mesh) == 0:
            self._mesh = guess_mesh(self, prefix="mesh")

        if not isinstance(self._mesh, Mesh):
            mesh_desc, *_ = group_dict_by_prefix(self._metadata, prefixes="mesh", sep="_")
            self._mesh = Mesh(self._mesh, parent=self, **(mesh_desc or {}))

        return self._mesh

    @property
    def domain(self) -> DomainBase:
        return self.mesh

    def __call__(self, *args, **kwargs) -> typing.Callable[..., ArrayType]:
        if all([isinstance(a, (array_type, float, int)) for a in args]):
            return self.__ppoly__()(*args, **kwargs)
        else:
            return super().__call__(*args, **kwargs)

    def grad(self, n=1) -> Field:
        ppoly = self.__functor__()

        if isinstance(ppoly, tuple):
            ppoly, opts = ppoly
        else:
            opts = {}

        if self.mesh.ndim == 2 and n == 1:
            return Field(
                (ppoly.partial_derivative(1, 0), ppoly.partial_derivative(0, 1)),
                mesh=self.mesh,
                name=f"\\nabla({self.__str__()})",
                **opts,
            )
        elif self.mesh.ndim == 3 and n == 1:
            return Field(
                (
                    ppoly.partial_derivative(1, 0, 0),
                    ppoly.partial_derivative(0, 1, 0),
                    ppoly.partial_derivative(0, 0, 1),
                ),
                mesh=self.mesh,
                name=f"\\nabla({self.__str__()})",
                **opts,
            )
        elif self.mesh.ndim == 2 and n == 2:
            return Field(
                (ppoly.partial_derivative(2, 0), ppoly.partial_derivative(0, 2), ppoly.partial_derivative(1, 1)),
                mesh=self.mesh,
                name=f"\\nabla^{n}({self.__str__()})",
                **opts,
            )
        else:
            raise NotImplemented(f"TODO: ndim={self.mesh.ndim} n={n}")

    def __array__(self) -> array_type:
        if self._cache is None or self._cache is _not_found_:
            if self._op is None:
                raise RuntimeError(f"op is not defined!")
            x = self.domain.points
            if len(x) == 0:
                raise RuntimeError(f"Domain is not defined!")
            
            self._cache = self._op(*x)
            
        return self._cache

    def __ppoly__(self):
        if self._ppoly is None:
            self._ppoly = self.domain.interpolate(self.__array__())
        return self._ppoly

    def derivative(self, order, *args, **kwargs) -> Field:
        if isinstance(order, int) and order < 0:
            func = self.__ppoly__().antiderivative(*order)
            return Field(func, mesh=self.mesh, label=f"I_{{{order}}}{{{self._render_latex_()}}}")
        elif isinstance(order, collections.abc.Sequence):
            func = self.__ppoly__().partial_derivative(*order)
            return Field(func, mesh=self.mesh, label=f"d_{{{order}}}{{{self._render_latex_()}}}")
        else:
            func = self.__ppoly__().derivative(order)
            return Field(func, mesh=self.mesh, label=f"d_{{{order}}}{{{self._render_latex_()}}}")

    def antiderivative(self, order: int, *args, **kwargs) -> Field:
        raise NotImplementedError(f"")

    def partial_derivative(self, order: typing.Tuple[int, ...], *args, **kwargs) -> Field:
        return self.derivative(order, *args, **kwargs)



import unittest
from copy import deepcopy

from spdm.core.Path import Path
from spdm.utils.logger import logger
from spdm.utils.tags import _not_found_
from spdm.utils.pluggable import Pluggable
import typing


class TBase(Pluggable):
    _registry = {}

    @classmethod
    def _guess_plugin_name(cls, cls_name, *args, **kwargs) -> typing.List[str]:
        return [cls_name]

 

@TBase.register("A")
class MyClassA:
    def my_method(self):
        print("Method called from MyClassA")


@TBase.register("B")
class MyClassB:
    def my_method(self):
        print("Method called from MyClassB")


class TestPlugin(unittest.TestCase):

    def test_create(self):
        foo = TBase("A")
        self.assertTrue(isinstance(foo, MyClassA))


if __name__ == '__main__':
    unittest.main()



import collections
import collections.abc
import typing
import pathlib
import f90nml
import numpy as np
from spdm.core.File import File
from spdm.core.Entry import Entry
from spdm.utils.tree_utils import normalize_data
from spdm.utils.logger import logger

# class NumpyEncoder(json.NAMELISTEncoder):
#     def default(self, obj):
#         if isinstance(obj, np.ndarray):
#             return obj.tolist()

#         logger.debug(type(obj))
# #         return super().default(obj)


# class NAMELISTFile(File):
#     def __init__(self, path, *args, **kwargs):
#         super().__init__(path, *args, **kwargs)
#         self._path = path

#     def load(self, *args, path=None,  **kwargs):
#         with open(path or self._path, mode="r") as fid:
#             self.root._holder = f90nml.read(fid).todict(complex_tuple=True)

#     def save(self,   *args, path=None, template=None, **kwargs):
#         with open(path or self._path, mode="w") as fid:
#             d = self._encode("", self.root._holder)
#             if template is None:
#                 f90nml.write(d, fid)
#             else:
#                 if isinstance(template, str):
#                     template = urisplit(template)
#                 f90nml.patch(template.path, d, fid)

#     def _encode(self, prefix, nobj):
#         if isinstance(nobj, str):
#             return nobj
#         elif isinstance(nobj, collections.abc.Mapping):
#             return {k: self._encode(f"{prefix}.{k}", p) for k, p in nobj.items()}
#         elif isinstance(nobj, collections.abc.Sequence):
#             return [self._encode(f"{prefix}.{k}", p) for k, p in enumerate(nobj)]
#         elif type(nobj) not in [str, int, float, bool, type(None)]:
#             return str(nobj)
#         else:
#             return nobj

@File.register(["namelist"])
class NAMELISTFile(File):
    def __init__(self,  *args, template: str = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._template = pathlib.Path(template) if template is not None else None

    def write(self, data=None, *args, **kwargs):
        if data is None:
            data = kwargs
        if not isinstance(data, collections.abc.Mapping) or self._template is None:
            super().write(data, *args, **kwargs)
        else:
            data = normalize_data(data)
            f90nml.patch(self._template.as_posix(), data, self.path.as_posix())

    def read(self) -> Entry:
        with open(self.path, mode="r") as fid:
            data: dict = f90nml.read(fid).todict(complex_tuple=True)
        return Entry(data)



import os
import pathlib
from spdm.core.File import File
from spdm.core.Entry import open_entry
from spdm.utils.logger import logger

WORKSPACE = "/home/salmon/workspace"  # "/ssd01/salmon_work/workspace/"
OUTPUT_PATH = f"{WORKSPACE}/output"

os.environ["SP_DATA_MAPPING_PATH"] = f"{WORKSPACE}/fytok_data/mapping"


if __name__ == '__main__':

    DATA_PATH = pathlib.Path(f"{WORKSPACE}/fytok_data/gfiles")

    # eq0 = open_entry(f"file+geqdsk:///{DATA_PATH.as_posix()}/g063982.04800", mode="r").fetch()

    # eq1 = open_entry(DATA_PATH/"g063982.04800", mode="r", format="geqdsk").fetch()

    # logger.debug(eq0)

    # logger.debug(eq1)

    # eq2 = open_entry(f"east+mdsplus://{WORKSPACE}/fytok_data/mdsplus/~t/?enable=efit_east", shot=70745)

    # # eq2 = open_entry(f"east+mdsplus://202.127.204.12", shot=70745)

    # logger.debug(eq2.child("equilibrium/time_slice/0/boundary/outline/r").fetch())

    # eq3 = open_entry(f"cfetr")

    # logger.debug(eq3.child("wall/description_2d/0/limiter/unit/0/outline/r").fetch())

    shot_num = 70754

    time_slice = 10

    entry = open_entry(f"east+mdsplus://202.127.204.12?enable=efit_east&shot={shot_num}")

    data = {
        "wall": entry.child(f"wall"),
        "equilibrium": {"time_slice": [entry.child(f"equilibrium/time_slice/{time_slice}")]}
    }

    with File(f"{OUTPUT_PATH}/g{shot_num}.gfile", mode="w", format="geqdsk") as fid:
        fid.write(data, description="equilibrium")



from spdm.utils.tree_utils import update_tree
from spdm.utils.logger import logger
import pprint
if __name__ == "__main__":

    d = update_tree(None, "a/b/c/1/2/c", "hello")

    pprint.pprint(d)

    update_tree(d, "a/b/d", {"hello": [1, 2, 3, 4, 5]})

    pprint.pprint(d)

    update_tree(d, None, {"a/b/d/hello": 789})

    pprint.pprint(d)



import collections.abc
import typing
import uuid

import bokeh
import bokeh.models.widgets as bk
import ipywidgets as ip
import jupyter_bokeh as jbk
import numpy as np
from bokeh.io import output_notebook
from bokeh.plotting import figure
from spdm.core.HTree import HTree
from spdm.geometry.BBox import BBox
from spdm.geometry.Circle import Circle
from spdm.geometry.Curve import Curve
from spdm.geometry.GeoObject import GeoObject
from spdm.geometry.Line import Line
from spdm.geometry.Point import Point
from spdm.geometry.PointSet import PointSet
from spdm.utils.logger import logger

from spdm.view.View import View


def bokeh_example():
    output_notebook()
    x = np.linspace(0, 2*np.pi, 2000)
    y = np.sin(x)
    p = figure(title="simple line example", width=600, height=300, y_range=(-5, 5),
               background_fill_color='#efefef')
    r = p.line(x, y, color="#8888cc", line_width=1.5, alpha=0.8)

    def update(f, w=1, A=1, phi=0):
        if f == "sin":
            func = np.sin
        elif f == "cos":
            func = np.cos
        else:
            func = np.sin
        r.data_source.data['y'] = A * func(w * x + phi)
    jbk.BokehModel(p)
    ip.interact(update, f=["sin", "cos"], w=(0, 50), A=(1, 10), phi=(0, 20, 0.1))


@View.register(["bokeh", "Bokeh"])
class BokehView(View):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def draw(self, obj, **kwargs) -> typing.Any:
        """
            Bokeh + Jupyter 生成可交互的widget
        """

        return obj



import typing
import unittest

import numpy as np
from scipy import constants
from spdm.utils.logger import logger
from spdm.geometry.GeoObject import GeoObject


class TestMesh(unittest.TestCase):

    def test_line(self):
        p0 = (0, 0)
        p1 = (1, 1)
        gobj = GeoObject(p0, p1, type="line")
        self.assertEqual(type(gobj).__name__, "Line")
        from spdm.geometry.Line import Line
        self.assertIsInstance(gobj, Line)

    def test_line2(self):
        from spdm.geometry.Line import Line
        p0 = (4, 5, 6)
        p1 = (1, 2, 3)
        gobj = Line(p0, p1)

        self.assertEqual(type(gobj).__name__, "Line")
        self.assertTrue(np.all(np.isclose(gobj.p0[:], p0)))
        self.assertTrue(np.all(np.isclose(gobj.p1[:], p1)))

    def test_coordinates(self):
        from spdm.geometry.Line import Line
        p0 = (4, 5, 6)
        p1 = (1, 2, 3)

        line = Line(p0, p1, coordinates="x,y,z")

        self.assertTrue(np.all(np.isclose(line.x, np.asarray([4, 1], dtype=float))))
        self.assertTrue(np.all(np.isclose(line.y, np.asarray([5, 2], dtype=float))))
        self.assertTrue(np.all(np.isclose(line.z, np.asarray([6, 3], dtype=float))))

    # def test_set(self):
    #     from spdm.geometry.Point import Point
    #     from spdm.geometry.GeoObject import GeoObjectSet
    #     gobj = GeoObjectSet(Point(1, 2, 3), Point(1, 2, 3))
    #     logger.debug(gobj.rank)
    #     logger.debug(len(gobj))


if __name__ == '__main__':
    unittest.main()



import abc
import typing

import numpy as np

from .GeoObject import GeoObject
from .Line import Line


class Plane(GeoObject):
    """ Plane
        平面，二维几何体
    """

    def __init__(self, *args,   **kwargs) -> None:
        super().__init__(*args, rank=2, **kwargs)

    @abc.abstractproperty
    def boundary(self) -> GeoObject:
        raise NotImplementedError(f"{self.__class__.__name__}")

    def map(self, u, *args, **kwargs):
        return NotImplemented

    def derivative(self, u, *args, **kwargs):
        return NotImplemented

    def dl(self, u, *args, **kwargs):
        return NotImplemented

    def pullback(self, func, *args, **kwargs):
        return NotImplemented

    def make_one_form(self, func):
        return NotImplemented



import collections
import collections.abc

import pathlib
import typing
import netCDF4 as nc
import numpy as np
from spdm.utils.logger import logger
from spdm.utils.tags import _undefined_, _not_found_
from spdm.core.Entry import Entry
from spdm.core.File import File
from spdm.core.Path import Path

SPDM_LIGHTDATA_MAX_LENGTH = 64


def nc_put_value(grp, path, value,  **kwargs):
    res = None
    path = Path(path)
    if isinstance(value, collections.abc.Mapping):
        for k, v in value.items():
            nc_put_value(grp, path/k,  v, **kwargs)
    elif isinstance(value, collections.abc.Sequence) and not isinstance(value, str):
        if all(map(lambda v: isinstance(v, (int, float)), value)):
            value = np.array(value)
            nc_put_value(grp, path, value, **kwargs)
        else:
            for k, v in enumerate(value):
                nc_put_value(grp, path/k, v)
    elif type(value) is np.ndarray and len(value) > SPDM_LIGHTDATA_MAX_LENGTH:
        parent = path.parent.__str__()
        key = path[-1]
        if len(parent) == 0:
            parent = grp
        else:
            parent = grp.createGroup(parent)

        dimensions = []
        for idx, d in enumerate(value.shape):
            parent.createDimension(f"{key}__dim_{idx}", d)
            dimensions.append(f"{key}__dim_{idx}")

        d = parent.createVariable(path.as_url(), value.dtype, tuple(dimensions))
        d[:] = value

    else:  # type(value) in [str, int, float]:
        p = path.parent.as_url()
        if len(p) > 0:
            obj = grp.createGroup(p)
        else:
            obj = grp
        
        obj.setncattr(path[-1], value)
    return


def nc_get_value(grp, path, projection=None, default=_not_found_, **kwargs):

    if grp is None:
        raise RuntimeError("None group")

    path = Path(path)[:]
    obj = grp
    for pos, p in enumerate(path):
        if isinstance(p, str):
            if p in obj.groups():
                obj = obj[p]
            elif p in obj.ncattrs() and pos == len(path)-1:
                obj = obj.getncattr(p)
            else:
                raise IndexError(path[:pos+1])

    if isinstance(obj, (nc.Group, nc.Dataset)):
        res1 = {k: nc_get_value(v, []) for k, v in obj.groups.items()}
        res2 = {k: nc_get_value(v, []) for k, v in obj.variables.items()}
        res3 = {k: obj.getncattr(k) for k in obj.ncattrs()}
        res = {**res1, **res2, **res3}
    elif isinstance(obj, nc.Variable):
        res = obj[:]
    else:
        res = obj

    return res


def nc_dump(grp):
    return nc_get_value(grp, [])


class NetCDFEntry(Entry):

    def __init__(self,   *args, **kwargs):
        super().__init__(*args, **kwargs)

    def copy(self, other):
        if hasattr(other, "__entry__"):
            other = other.__entry__.__value__()
        self.update(other)

    def insert(self,  value, *args, **kwargs):
        return nc_put_value(self._data, self._path, value, *args, **kwargs)

    def find(self,   *args, **kwargs) -> typing.Any:
        return nc_get_value(self._data, self._path, *args, **kwargs)

    def dump(self):
        return nc_dump(self._data)

    def iter(self,  path, *args, **kwargs):
        raise NotImplementedError()


@File.register(["nc", "netcdf", "NetCDf"])
class FILEPLUGINnetcdf(File):

    MOD_MAP = {File.Mode.read: "r",
               File.Mode.read | File.Mode.write: "r+",
               File.Mode.write: "w-",
               File.Mode.write | File.Mode.create: "w",
               File.Mode.read | File.Mode.write | File.Mode.create: "a",
               }

    """
        r       Readonly, file must exist (default)
        r+      Read/write, file must exist
        w       Create file, truncate if exists
        w- or x Create file, fail if exists
        a       Read/write if exists, create otherwise
    """

    def __init__(self,  *args,  **kwargs):
        super().__init__(*args,   **kwargs)
        self._fid = None

    @property
    def mode_str(self) -> str:
        return FILEPLUGINnetcdf.MOD_MAP[self.mode]

    def open(self) -> File:
        if self.is_open:
            return self

        try:
            if self._fid is None:
                self._fid = nc.Dataset(self.path,  self.mode_str, format="NETCDF4")
        except OSError as error:
            raise FileExistsError(f"Can not open file {self.path}! {error}")
        else:
            logger.debug(f"Open NetCDF File {self.path} mode={self.mode}")

        super().open()

        return self

    def close(self):
        if not self.is_open:
            return
        if self._fid is not None:
            self._fid.close()
        self._fid = None
        return super().close()

    def read(self, lazy=True) -> Entry:
        return NetCDFEntry(self.open()._fid)

    def write(self, *args, **kwargs):
        NetCDFEntry(self.open()._fid).insert(*args, **kwargs)


# class NetCDFCollection(FileCollection):
#     def __init__(self, uri, *args, **kwargs):
#         super().__init__(uri, *args,
#                          file_extension=".h5",
#                          file_factory=lambda *a, **k: NCFile(*a, **k),
#                          ** kwargs)

__SP_EXPORT__ = FILEPLUGINnetcdf



from __future__ import annotations

import collections.abc
import typing
import uuid
from copy import copy
from functools import cached_property

import numpy as np

from ..core.HTree import List
from ..utils.logger import logger
from ..utils.typing import ArrayLike, ArrayType, NumericType, ScalarType, array_type, nTupleType, numeric_type


class BBox:
    def __init__(self, origin: ArrayLike, dimensions: ArrayLike, transform=None, shift=None) -> None:
        self._origin = np.asarray(origin)
        self._dimensions = np.asarray(dimensions)
        self._transform = transform
        self._shift = shift

    def __copy__(self) -> BBox:
        return BBox(self._origin, self._dimensions, self._transform, self._shift)

    def __repr__(self) -> str:
        """x y width height"""
        return f"viewBox=\"{' '.join([*map(str,self._origin)])}  {' '.join([*map(str,self._dimensions)]) }\" transform=\"{self._transform}\" shift=\"{self._shift}\""

    @property
    def origin(self) -> ArrayType:
        return self._origin

    @property
    def dimensions(self) -> ArrayType:
        return self._dimensions

    @property
    def is_valid(self) -> bool:
        return np.all(self._dimensions > 0) == True

    @property
    def is_degraded(self) -> bool:
        return (np.any(np.isclose(self._dimensions, 0.0))) == True

    # def __equal__(self, other: BBox) -> bool:
    #     return np.allclose(self._xmin, other._xmin) and np.allclose(self._xmax, other._xmax)

    # def __or__(self, other: BBox) -> BBox:
    #     if other is None:
    #         return self
    #     else:
    #         return BBox(np.min(self._xmin, other._xmin), np.max(self._xmax, other._xmax))

    # def __and__(self, other: BBox) -> BBox | None:
    #     if other is None:
    #         return None
    #     else:
    #         res = BBox(np.max(self._xmin, other._xmin), np.min(self._xmax, other._xmax))
    #         return res if res.is_valid else None

    @property
    def ndim(self) -> int:
        return len(self._dimensions)

    @property
    def center(self) -> ArrayType:
        return self._origin + self._dimensions * 0.5

    """ center of geometry """

    @property
    def measure(self) -> float:
        return float(np.product(self._dimensions))

    """ measure of geometry, length,area,volume,etc. 默认为 bbox 的体积 """

    def enclose(self, *args) -> bool | array_type:
        """Return True if all args are inside the geometry, False otherwise."""

        if len(args) == 1:

            # if hasattr(args[0], "bbox"):
            #     return self.enclose(args[0].bbox)
            # elif isinstance(args[0], BBox):
            #     return self.enclose(args[0].origin) and self.enclose(args[0].origin+args[0].dimensions)
            if hasattr(args[0], "points"):
                return self.enclose(*args[0].points)
            if isinstance(args[0], collections.abc.Sequence):
                return self.enclose(*args[0])
            elif isinstance(args[0], array_type):
                return self.enclose([args[0][..., idx] for idx in range(self.ndim)])
            else:
                raise TypeError(f"args has wrong type {type(args[0])} {args}")

        elif len(args) == self.ndim:
            if isinstance(args[0], array_type):
                r_pos = [args[idx] - self._origin[idx] for idx in range(self.ndim)]
                return np.bitwise_and.reduce(
                    [((r_pos[idx] >= 0) & (r_pos[idx] <= self._dimensions[idx])) for idx in range(self.ndim)]
                )
            else:
                res = all(
                    [
                        ((args[idx] >= self._origin[idx]) and (args[idx] <= self._origin[idx] + self._dimensions[idx]))
                        for idx in range(self.ndim)
                    ]
                )
                if not res:
                    logger.debug((args, self._origin, self._dimensions))
                return res

        else:
            raise TypeError(f"args has wrong type {type(args[0])} {args}")

    def union(self, other: BBox) -> BBox:
        raise NotImplementedError(f"intersection")

    """ Return the union of self with other. """

    def intersection(self, other: BBox):
        raise NotImplementedError(f"intersection")

    """ Return the intersection of self with other. """

    def reflect(self, point0, pointt1):
        raise NotImplementedError(f"reflect")

    """ reflect  by line"""

    def rotate(self, angle, axis=None):
        raise NotImplementedError(f"rotate")

    """ rotate  by angle and axis"""

    def scale(self, *s, point=None):
        raise NotImplementedError(f"scale")

    """ scale self by *s, point """

    def translate(self, *shift):
        raise NotImplementedError(f"translate")



import abc
import collections
import collections.abc
import inspect
import os
import sys
from enum import Enum
import typing

from .logger import logger
from .misc import camel_to_snake
from .sp_export import sp_find_module, sp_load_module, walk_namespace_modules
from .tags import _not_found_


class Pluggable(metaclass=abc.ABCMeta):
    """Factory class to create objects from a registry."""

    _plugin_prefix = __package__
    _plugin_registry = {}

    @classmethod
    def _get_plugin_fullname(cls, name: str) -> str:
        prefix = getattr(cls, "_plugin_prefix", None)
        if prefix is None:
            name = name.replace("/", ".").lower()
            m_pth = camel_to_snake(cls.__module__).split(".")
            prefix = ".".join(m_pth[0:1] + ["plugins"] + m_pth[1:] + [""]).lower()
            cls._plugin_prefix = prefix
        if not name.startswith(prefix):
            name = prefix + name
        return name

    @classmethod
    def register(cls, sub_list: str | list | None = None, plugin_cls=None):
        """
        Decorator to register a class to the registry.
        """
        if plugin_cls is not None:
            if not isinstance(sub_list, list):
                sub_list = [sub_list]

            for name in sub_list:
                if not isinstance(name, str):
                    continue
                cls._plugin_registry[cls._get_plugin_fullname(name)] = plugin_cls

            return plugin_cls
        else:

            def decorator(o_cls):
                cls.register(sub_list, o_cls)
                return o_cls

            return decorator

    @classmethod
    def _plugin_guess_name(cls, self, *args, **kwargs) -> str | None:
        return kwargs.pop("plugin_name", None)

    @classmethod
    def __dispatch_init__(cls, sub_list, self, *args, **kwargs) -> None:
        if sub_list is None:
            sub_list = cls._plugin_guess_name(self, *args, **kwargs)

        if sub_list is None:
            sub_list = []
        elif not isinstance(sub_list, list):
            sub_list = [sub_list]

        n_cls = None

        for sub_cls in sub_list:
            if sub_cls is None or sub_cls is _not_found_ or sub_cls == "":
                continue

            if isinstance(sub_cls, Enum):
                sub_cls = sub_cls.name

            if inspect.isclass(sub_cls):
                n_cls = sub_cls
                sub_cls = n_cls.__name__
            elif not isinstance(sub_cls, str):
                logger.warning(f"Invalid plugin name {sub_cls}!")
                continue
            elif sub_cls == "dummy":
                n_cls = cls
                break
            else:
                cls_name = cls._get_plugin_fullname(sub_cls)

                if cls_name not in cls._plugin_registry:
                    sp_load_module(cls_name)

                n_cls = cls._plugin_registry.get(cls_name, None)

            if n_cls is not None:
                break

        if (n_cls is cls or n_cls is None) and ("dummy" in sub_list or len(sub_list) == 0):
            return
        elif inspect.isclass(n_cls) and issubclass(n_cls, cls):
            self.__class__ = n_cls
            n_cls.__init__(self, *args, **kwargs)

        else:
            raise ModuleNotFoundError(
                f"Can not find module as subclass of '{cls.__name__}' {n_cls} from {sub_list} in {cls._plugin_registry}!"
            )

    def __init__(self, *args, **kwargs) -> None:
        if self.__class__ is Pluggable:
            self.__class__.__dispatch_init__(None, self, *args, **kwargs)
            return

    @classmethod
    def _find_plugins(cls) -> typing.List[str]:
        """Find all plugins in the Python path.

        仅返回可的module （import 成功的）
        """
        head = len(cls._plugin_prefix)
        return [p[head:] for p in walk_namespace_modules(cls._plugin_prefix[:-1])]



import collections.abc
import os
import pprint
import typing
from ..core.Expression import Variable
import numpy as np
import scipy.optimize
import scipy.ndimage  # for maximum_filter,binary_erosion, generate_binary_structure

from ..core.Field import Field
from ..utils.logger import logger
from ..utils.typing import ArrayType, NumericType, ScalarType

SP_EXPERIMENTAL = os.environ.get("SP_EXPERIMENTAL", False)

# logger.info(f"SP_EXPERIMENTAL \t: {SP_EXPERIMENTAL}")

EPSILON = 1.0e-2


def minimize_filter(func: typing.Callable[..., ScalarType | ArrayType], X, Y, width=None,
                    tolerance: float = None,
                    method="L-BFGS-B"
                    # xmin: float, ymin: float, xmax: float, ymax: float, tolerance: float = EPSILON
                    ):

    # if isinstance(tolerance, float):
    #     dx = tolerance
    #     dy = tolerance
    # elif isinstance(tolerance, (collections.abc.Sequence, np.ndarray)) and len(tolerance) == 2:
    #     dx, dy = tolerance
    # else:
    #     raise TypeError(f"Illegal type {type(dx)}")

    # nx = int((xmax-xmin)/dx)+1
    # ny = int((ymax-ymin)/dy)+1

    # X, Y = np.meshgrid(np.linspace(xmin, xmax, nx),
    #                    np.linspace(ymin, ymax, ny), indexing='ij')

    data = func(X, Y)

    nx, ny = data.shape

    z_max = np.max(data)

    z_min = np.min(data)

    if tolerance is None:
        tolerance = np.abs(z_max-z_min)/np.sqrt(nx*ny)*2
        if tolerance > 0.01 or tolerance < 0.001:
            tolerance = 0.01

    if isinstance(width, collections.abc.Sequence):
        wx, wy = width
    else:
        wx = int(max(4, nx/32))
        wy = int(max(4, ny/32))

    peak = scipy.ndimage.minimum_filter(data, size=(wx, wy), mode='constant') == data

    idxs = np.asarray(np.where(peak)).T

    for ix, iy in idxs:

        if ix == 0 or iy == 0 or ix == nx-1 or iy == ny-1:
            continue

        r = np.abs((data[ix, iy]-z_min)/(z_max-z_min))

        if r > tolerance:
            continue

        xmin = X[ix-1, iy]
        xmax = X[ix+1, iy]
        ymin = Y[ix, iy-1]
        ymax = Y[ix, iy+1]

        x = X[ix, iy]
        y = Y[ix, iy]

        # if True:

        sol = scipy.optimize.minimize(lambda x: func(x[0], x[1]), np.asarray([x, y]),
                                      bounds=[(xmin, xmax), (ymin, ymax)],
                                      method=method,
                                      tol=tolerance)

        xsol, ysol = sol.x

        if not (x >= xmin or x <= xmax or y >= ymin or y <= ymax):
            continue
        elif sol.success:
            yield xsol, ysol
        else:
            logger.warning(f"{sol.message} at {xsol, ysol} ")
            yield x, y




from __future__ import annotations

import collections
import collections.abc
import typing
from copy import copy
from functools import cached_property

import numpy as np
import scipy.constants
from scipy.interpolate import CubicSpline, PPoly
from spdm.geometry.GeoObject import GeoObject

from ..utils.logger import logger
from ..utils.typing import ArrayType, array_type, nTupleType
from .BBox import BBox
from .Curve import Curve

TWOPI = 2.0*scipy.constants.pi


@Curve.register("cubic_spline_curve")
class CubicSplineCurve(Curve):
    def __init__(self, *args, uv=None,  **kwargs) -> None:

        super().__init__(*args, **kwargs)
        self._points = points
        self._uv = uv if uv is None else np.linspace(0, 1, self._points.shape[0])

    def __copy__(self) -> CubicSplineCurve:
        other: CubicSplineCurve = super().__copy__()  # type:ignore
        other._points = self._points
        other._uv = self._uv
        return other



    def coordinates(self, *uvw, **kwargs) -> ArrayType:
        if len(uvw) == 0:
            return self._points
        else:
            return self._spl(*uvw, **kwargs)

    @cached_property
    def _spl(self) -> PPoly:
        return CubicSpline(self._uv, self._points, bc_type="periodic" if self.is_closed else "not-a-knot")

    @cached_property
    def _derivative(self):
        return self._spl.derivative()

    def derivative(self, *args, **kwargs):
        if len(args) == 0:
            args = [self._uv]
        res = self._derivative(*args, **kwargs)
        return res[:, 0], res[:, 1]

    @cached_property
    def bbox(self) -> BBox:
        """ bbox of geometry """
        points = self._points
        return BBox([np.min(x) for x in points], [np.max(x) for x in points])

    def enclose(self, *args) -> bool:
        if not self.is_closed:
            return False
        return super().enclose(*args)

    def remesh(self, u) -> CubicSplineCurve:
        other = copy(self)
        if isinstance(u, array_type):
            other._uv = u
        elif callable(u):
            other._uv = u(*self.points)
        else:
            raise TypeError(f"illegal type u={type(u)}")
        return other




from .Surface import Surface
from .Solid import Solid
from .Plane import Plane
from .Curve import Curve
from .Circle import Circle


@Surface.register("toroidal_surface")
class ToroidalSurface(Surface):
    def __init__(self, cross_section: Curve, circle: Circle, **kwargs) -> None:
        super().__init__(**kwargs)


@Surface.register("toroidal")
class Toroidal(Solid):
    def __init__(self, cross_section: Plane, circle: Circle, *args, **kwargs) -> None:
        super().__init__(**kwargs)
        self._boundary = ToroidalSurface(cross_section.boundary, circle, *args,)



import collections.abc
import dataclasses
import inspect
import typing
from dataclasses import dataclass
from enum import Enum
import types
import numpy as np
import numpy.typing as np_tp

from .logger import logger, deprecated
from .tags import _not_found_

ArrayLike = np_tp.ArrayLike | None

boolean_type = (bool,)

integral_type = (
    int,
    np.integer,
)

real_type = (
    float,
    np.floating,
)

bitwise_and_reduce = np.bitwise_and.reduce

complex_type = (
    complex,
    np.complexfloating,
)

ScalarType = (
    bool
    | int
    | float
    | complex
    | np.float64
    | np.complex64
    | np.complex128
    | np.integer
    | np.floating
    | np.bool_
    | None
)

scalar_type = (*boolean_type, *integral_type, *real_type, *complex_type)

ArrayType = np_tp.NDArray[np.floating | np.complexfloating]

array_type = np.ndarray


def array_like(x: array_type, value) -> array_type:
    if callable(value):
        return value(x)
    elif isinstance(value, array_type) and x.size == value.size:
        return value
    elif value is _not_found_:
        return np.full_like(x, 0)
    else:
        return np.full_like(x, value)


NumericType = ScalarType | ArrayType

numeric_type = (*scalar_type, array_type)

PrimaryType = str | NumericType | Enum

primary_type = (str, *numeric_type)

NativeType = PrimaryType | dict | list

native_type = (dict, list, *primary_type)

_T = typing.TypeVar("_T")

HTContainer = _T | typing.Sequence[_T]
"""Hierarchical Container Type  """

HNodeLike = None | int | str | bool | ArrayType

HTreeLike = dict | list | HNodeLike


@dataclass
class Vector2:
    x: float
    y: float


@dataclass
class Vector3:
    x: float
    y: float
    z: float


@dataclass
class Vector4:
    x: float
    y: float
    z: float
    t: float


nTupleType = typing.Tuple[ScalarType, ...]


def is_vector(v: typing.Any) -> bool:
    return isinstance(v, collections.abc.Sequence) and all(is_scalar(d) for d in v)


def is_array(v: typing.Any) -> bool:
    return isinstance(v, array_type)


def is_numeric(v: typing.Any) -> bool:
    return isinstance(v, numeric_type)


def is_int(v: str | int):
    if isinstance(v, int):
        return v
    elif isinstance(v, str):
        try:
            v = int(v)
        except ValueError:
            return False
        else:
            return v
    else:
        return False


def is_complex(d: typing.Any) -> bool:
    return np.iscomplexobj(d)


def is_real(d: typing.Any) -> bool:
    return not np.iscomplexobj(d)


def is_tree(d) -> bool:
    return isinstance(d, (collections.abc.Mapping, collections.abc.Sequence)) and not isinstance(d, str)


# def is_scalar(d: typing.Any) -> bool:
#     return isinstance(d, (int, float, complex, np.floating, np.complexfloating)) or hasattr(d.__class__, "__float__")
is_scalar = np.isscalar


def as_scalar(d: typing.Any) -> ScalarType:
    return complex(d) if is_complex(d) else float(d)


def as_array(d: typing.Any, *args, **kwargs) -> ArrayType:
    if d is None or d is _not_found_:
        return d
    elif isinstance(d, array_type):
        return d
    elif hasattr(d, "__array__"):
        return d.__array__()
    elif hasattr(d.__class__, "_value_"):
        return np.asarray(d.__value__, *args, **kwargs)
    else:
        return np.asarray(d, *args, **kwargs)


def normalize_array(value: ArrayLike, *args, **kwargs) -> ArrayType:
    """将 value 转换为 array_type 类型"""
    if isinstance(value, array_type) or is_scalar(value):
        pass
    elif value is None or value is _not_found_:
        value = None
    elif isinstance(value, numeric_type):
        value = as_array(value, *args, **kwargs)
    elif isinstance(value, tuple):
        value = as_array(value, *args, **kwargs)
    elif isinstance(value, collections.abc.Sequence):
        value = as_array(value, *args, **kwargs)
    elif isinstance(value, collections.abc.Mapping) and len(value) == 0:
        value = None
    else:
        raise RuntimeError(f"Function._normalize_value() incorrect value {value}! {type(value)}")

    if isinstance(value, array_type) and value.size == 1:
        value = np.squeeze(value).item()

    return value


def as_dataclass(cls, value):
    if not dataclasses.is_dataclass(cls):
        raise TypeError(type(cls))
    elif isinstance(value, collections.abc.Mapping):
        value = cls(**{k: value.get(k, None) for k in cls.__dataclass_fields__})
    elif isinstance(value, collections.abc.Sequence) and not isinstance(value, str):
        value = cls(*value)
    else:
        raise TypeError(f"Can not convert '{type(value)}' to dataclass")
        # value = type_hint(value, **kwargs)
    return value


def as_namedtuple(d: dict, name=""):
    return collections.namedtuple(name, d.keys())(d.values())


def as_value(obj: typing.Any) -> HTreeLike:
    if obj is _not_found_ or obj is None:
        return obj
    elif isinstance(obj, collections.abc.Mapping):
        return {k: as_value(v) for k, v in obj.items()}
    elif isinstance(obj, collections.abc.Sequence) and not isinstance(obj, str):
        return [as_value(v) for v in obj]
    elif hasattr(obj.__class__, "_value_"):
        return obj.__value__
    else:
        return obj


def convert_to_named_tuple(d=None, ntuple=None, **kwargs):
    if d is None and len(kwargs) > 0:
        d = kwargs
    if d is None:
        return d
    elif hasattr(ntuple, "_fields") and isinstance(ntuple, type):
        return ntuple(*[try_get(d, k) for k in ntuple._fields])
    elif isinstance(d, collections.abc.Mapping):
        keys = [k.replace("$", "s_") for k in d.keys()]
        values = [convert_to_named_tuple(v) for v in d.values()]
        if not isinstance(ntuple, str):
            ntuple = "__" + ("_".join(keys))
        ntuple = ntuple.replace("$", "_")
        return collections.namedtuple(ntuple, keys)(*values)
    elif isinstance(d, collections.abc.MutableSequence):
        return [convert_to_named_tuple(v) for v in d]
    else:
        return d


def get_origin(tp: typing.Any) -> typing.Type:
    """
    获得 object，Type，typing.Generic 的原始类型
    """

    if isinstance(tp, typing.types.UnionType):
        return get_origin(typing.get_args(tp)[0])

    orig_class = typing.get_origin(tp)

    if orig_class is not None:
        return orig_class
    elif inspect.isclass(tp):
        return tp
    else:
        return get_origin(getattr(tp, "__orig_class__", tp.__class__))
        # raise TypeError(f"{tp} must be a class or GenericAlias! {type(tp)}")

    # elif hasattr(tp, "__orig_bases__"):
    #     orig_bases = getattr(tp, "__orig_bases__", None)

    #     if orig_bases is None:
    #         return tp
    #     elif isinstance(orig_bases, tuple):
    #         return get_origin(orig_bases[0])
    #     else:
    #         return get_origin(orig_bases)
    # elif inspect.isclass(tp):
    #     return tp


def get_orig_bases(tp: typing.Any) -> typing.Tuple[typing.Type, ...]:
    return getattr(tp, "__orig_bases__", tuple([]))


def get_args(tp: typing.Any) -> typing.Tuple[typing.Type, ...]:
    """
    获得 typing.Generic 类型的 type_hint
    例如： 若定义泛型类
        class Foo[typing.Genric[_T]]:
            pass
        obj=Foo[int]
        get_generic_args(obj)

        >>> int

    """
    if tp is None or tp is _not_found_:
        res = tuple()

    elif isinstance(tp, tuple):
        res = sum([get_args(t) for t in tp], tuple())

    elif typing.get_origin(tp) is not None:
        res = typing.get_args(tp)

    elif len(get_orig_bases(tp)) > 0:
        return sum([get_args(t) for t in get_orig_bases(tp)], tuple())

    else:
        return get_args(getattr(tp, "__orig_class__", tp.__class__))

    return tuple([t for t in res if t is not None and not isinstance(t, typing.TypeVar)])


def get_type(obj):
    return getattr(obj, "__orig_class__", obj.__class__)


def get_type_hint(tp: typing.Type | types.GenericAlias, prop_name: str):
    if not isinstance(tp, (typing.Type, types.GenericAlias)):
        tp = getattr(tp, "__orig_class__", tp.__class__)

    return typing.get_type_hints(tp).get(prop_name, None)


_T = typing.TypeVar("_T")


def isinstance_generic(obj: typing.Any, type_hint: typing.Type) -> bool:
    """判断 obj 是否是 type_hint 的实例,
    type_hint 可以是 typing.GenericAlias 或者 typing.Type
    """
    if type_hint is None:
        return False

    elif type_hint is _not_found_ or type_hint is typing.Any:
        return True

    elif isinstance(type_hint, typing.types.UnionType):
        return any([isinstance_generic(obj, tp) for tp in typing.get_args(type_hint)])

    elif inspect.isclass(type_hint):
        return isinstance(obj, type_hint)

    orig_class = typing.get_origin(type_hint)

    if inspect.isclass(type_hint) and orig_class is None:
        return isinstance(obj, type_hint)

    elif inspect.isclass(type_hint) and obj.__class__ == type_hint:
        return True

    elif orig_class is None:
        # raise RuntimeError(type_hint)
        return False

    elif not isinstance(obj, orig_class):
        return False

    elif getattr(obj, "__orig_class__", obj.__class__) == type_hint:
        return True

    elif not hasattr(obj, "__orig_class__"):
        return True

    elif type_hint in getattr(obj, "__orig_bases__", []):
        return True

    else:
        return False


def type_convert(tp: typing.Type, value: typing.Any, *args, **kwargs) -> typing.Any:
    if value is _not_found_:
        value = kwargs.pop("default_value", _not_found_)

    if tp is None or tp is _not_found_ or tp is typing.Any:
        return value

    elif isinstance(tp, typing.types.UnionType):
        return type_convert(typing.get_args(tp)[0], value, *args, **kwargs)

    elif isinstance_generic(value, tp):
        return value

    elif tp in (set, list, dict, tuple):
        return tp(value)

    elif issubclass(get_origin(tp), array_type):
        value = as_array(value)

    elif tp in primary_type:
        value = getattr(value, "__value__", value)

        if value is not _not_found_:
            try:
                tmp = tp(value)
            except Exception as error:
                raise TypeError(f"Can not convert {value} to {tp}") from error
            else:
                value = tmp

    elif dataclasses.is_dataclass(tp):
        value = getattr(value, "__value__", value)
        value = as_dataclass(tp, value)

    elif issubclass(get_origin(tp), Enum):
        value = getattr(value, "__value__", value)

        if isinstance(value, collections.abc.Mapping):
            value = tp[value["name"]]

        elif isinstance(value, (int, str)):
            value = tp[value]

        else:
            raise TypeError(f"Can not convert {value} to {tp}")

    else:
        try:
            value = tp(value, *args, **kwargs)
        except Exception as error:
            raise TypeError(f"Can not convert {type(value)} to {tp}") from error

    return value


def convert_to_named_tuple(d=None, ntuple=None, **kwargs):
    if d is None and len(kwargs) > 0:
        d = kwargs
    if d is None:
        return d
    elif hasattr(ntuple, "_fields") and isinstance(ntuple, type):
        return ntuple(*[try_get(d, k) for k in ntuple._fields])
    elif isinstance(d, collections.abc.Mapping):
        keys = [k.replace("$", "s_") for k in d.keys()]
        values = [convert_to_named_tuple(v) for v in d.values()]
        if not isinstance(ntuple, str):
            ntuple = "__" + ("_".join(keys))
        ntuple = ntuple.replace("$", "_")
        return collections.namedtuple(ntuple, keys)(*values)
    elif isinstance(d, collections.abc.MutableSequence):
        return [convert_to_named_tuple(v) for v in d]
    else:
        return d


def as_native(d, enable_ndarray=True) -> typing.Union[str, bool, float, int, np.ndarray, dict, list]:
    """
    convert d to native data type str,bool,float, int, dict, list
    """
    if isinstance(d, (bool, int, float, str)):
        return d
    elif isinstance(d, np.ndarray):
        return d.tolist() if not enable_ndarray else d
    elif isinstance(d, collections.abc.Mapping):
        return {as_native(k): as_native(v, enable_ndarray=enable_ndarray) for k, v in d.items()}
    elif isinstance(d, collections.abc.Sequence):
        return [as_native(v, enable_ndarray=enable_ndarray) for v in d]

    else:
        logger.debug(type(d))
        return str(d)


def serialize(obj) -> typing.Any:
    """将对象 object serialize 为 native tree结构 （dict,list,np.ndarray,str,bool,int,float)"""

    if obj is None or isinstance(obj, primary_type):
        return obj
    elif isinstance(obj, collections.abc.Mapping):
        return {serialize(k): serialize(v) for k, v in obj.items()}
    elif isinstance(obj, collections.abc.Sequence):
        return [serialize(v) for v in obj]
    elif hasattr(obj, "__serialize__"):
        return obj.__serialize__()
    else:
        raise TypeError(f"Not support type: {type(obj)}")


def dump(obj) -> typing.Any:
    if obj is None or isinstance(obj, primary_type):
        return obj
    elif isinstance(obj, collections.abc.Mapping):
        return {dump(k): dump(v) for k, v in obj.items()}
    elif isinstance(obj, collections.abc.Sequence):
        return [dump(v) for v in obj]
    elif hasattr(obj, "_value_"):
        return dump(obj.__value__)
    else:
        raise TypeError(f"Not support type: {type(obj)}")





import numpy as np

def sp_read_geqdsk(file):
    """
    :param file: input file / file path
    :return: profile object
    """

    description = file.read(48)
    idum = int(file.read(4))
    nw = int(file.read(4))
    nh = int(file.read(4))
    file.readline()

    rdim = float(file.read(16))
    zdim = float(file.read(16))
    rcentr = float(file.read(16))
    rleft = float(file.read(16))
    zmid = float(file.read(16))
    file.readline()
    rmaxis = float(file.read(16))
    zmaxis = float(file.read(16))
    simag = float(file.read(16))
    sibry = float(file.read(16))
    bcentr = float(file.read(16))
    file.readline()
    current = float(file.read(16))
    simag = float(file.read(16))
    xdum = float(file.read(16))
    rmaxis = float(file.read(16))
    xdum = float(file.read(16))
    file.readline()

    zmaxis = float(file.read(16))
    xdum = float(file.read(16))
    sibry = float(file.read(16))
    xdum = float(file.read(16))
    xdum = float(file.read(16))
    file.readline()

    def _read_data(count, width=16):
        data = []
        for n in range(count):
            data.append(float(file.read(width)))
            if n >= count - 1 or ((n + 1) % 5 == 0):
                file.readline()
        data = np.asarray(data)
        return data

    #
    fpol = _read_data(nw)
    pres = _read_data(nw)
    ffprim = _read_data(nw)
    pprim = _read_data(nw)

    psirz = _read_data(nw * nh).reshape([nh, nw])

    qpsi = _read_data(nw)

    try:
        nbbs = int(file.read(5))
        limitr = int(file.read(5))
        file.readline()

        bbsrz = _read_data(nbbs * 2).reshape([nbbs, 2])
        limrz = _read_data(limitr * 2).reshape([limitr, 2])
    except:
        nbbs = 0
        limitr = 0
        bbsrz = None
        limrz = None

    data = {
        "description": description,
        # "idum": idum,
        "nw": nw,
        "nh": nh,
        "rdim": rdim,
        "zdim": zdim,
        "rcentr": rcentr,
        "rleft": rleft,
        "zmid": zmid,
        "rmaxis": rmaxis,
        "zmaxis": zmaxis,
        "simag": simag,
        "sibry": sibry,
        "bcentr": bcentr,
        "current": current,
        # "simag": simag,
        # "rmaxis": rmaxis,
        # "zmaxis": zmaxis,
        # "sibry": sibry,
        "fpol": fpol,
        "pres": pres,
        "ffprim": ffprim,
        "pprim": pprim,
        "psirz": psirz,
        "qpsi": qpsi,
        "bbsrz": bbsrz,
        "limrz": limrz,

    }
    return data


def sp_write_geqdsk(p, file):
    """
    :param profile: object

    :param file: file path / file
    :return:
    """

    nw = p["nw"]
    nh = p["nh"]

    file.write("%48s%4i%4i%4i\n" % (p["description"], 3, p["nw"], p["nh"]))
    file.write("%16.9e%16.9e%16.9e%16.9e%16.9e\n" %
               (p["rdim"], p["zdim"], p["rcentr"], p["rleft"], p["zmid"]))
    file.write("%16.9e%16.9e%16.9e%16.9e%16.9e\n" %
               (p["rmaxis"], p["zmaxis"], p["simag"], p["sibry"], p["bcentr"]))
    file.write("%16.9e%16.9e%16.9e%16.9e%16.9e\n" %
               (p["current"], p["simag"], 0, p["rmaxis"], 0))
    file.write("%16.9e%16.9e%16.9e%16.9e%16.9e\n" %
               (p["zmaxis"], 0, p["sibry"], 0, 0))

    def _write_data(d):
        count = len(d)
        for n in range(count):
            file.write("%16.9e" % d[n])
            if (n == count - 1) or ((n + 1) % 5 == 0):
                file.write('\n')

    _write_data(p["fpol"])
    _write_data(p["pres"])
    _write_data(p["ffprim"])
    _write_data(p["pprim"])
    _write_data(p["psirz"].reshape([nw * nh]))
    _write_data(p["qpsi"])
    file.write("%5i%5i\n" % (p["bbsrz"].shape[0], p["limrz"].shape[0]))
    _write_data(p["bbsrz"].reshape([p["bbsrz"].size]))
    _write_data(p["limrz"].reshape([p["limrz"].size]))

    return





import numpy as np
import imas
import pprint


def sp_read_geqdsk(file):
    """

    :param file: input file / file path
    :return: profile object
    """
    if type(file) is str:
        file = open(file, "r")

    # res = np.genfromtxt(file, dtype=np.dtype([
    #     ("description", 'S50'), ("idum", "i4"), ("nw", "i4"), ("nh", "i4")
    # ]))

    description = file.read(48)
    idum = int(file.read(4))
    nw = int(file.read(4))
    nh = int(file.read(4))
    file.readline()

    rdim = float(file.read(16))
    zdim = float(file.read(16))
    rcentr = float(file.read(16))
    rleft = float(file.read(16))
    zmid = float(file.read(16))
    file.readline()
    rmaxis = float(file.read(16))
    zmaxis = float(file.read(16))
    simag = float(file.read(16))
    sibry = float(file.read(16))
    bcentr = float(file.read(16))
    file.readline()
    current = float(file.read(16))
    simag = float(file.read(16))
    xdum = float(file.read(16))
    rmaxis = float(file.read(16))
    xdum = float(file.read(16))
    file.readline()

    zmaxis = float(file.read(16))
    xdum = float(file.read(16))
    sibry = float(file.read(16))
    xdum = float(file.read(16))
    xdum = float(file.read(16))
    file.readline()

    def _read_data(count, width=16):
        data = np.ndarray(shape=[count], dtype=float)

        for n in range(count):
            data[n] = float(file.read(width))
            if n >= count - 1 or ((n + 1) % 5 == 0):
                file.readline()
        return data

    #
    fpol = _read_data(nw)
    pres = _read_data(nw)
    ffprim = _read_data(nw)
    pprim = _read_data(nw)

    psirz = _read_data(nw * nh).reshape([nw, nh])

    qpsi = _read_data(nw)

    nbbs = int(file.read(5))
    limitr = int(file.read(5))
    file.readline()

    bbsrz = _read_data(nbbs * 2).reshape([nbbs, 2])
    limrz = _read_data(limitr * 2).reshape([limitr, 2])
    file.close()
    return {
        "description": description,
        # "idum": idum,
        "nw": nw,
        "nh": nh,
        "rdim": rdim,
        "zdim": zdim,
        "rcentr": rcentr,
        "rleft": rleft,
        "zmid": zmid,
        "rmaxis": rmaxis,
        "zmaxis": zmaxis,
        "simag": simag,
        "sibry": sibry,
        "bcentr": bcentr,
        "current": current,
        # "simag": simag,
        # "rmaxis": rmaxis,
        # "zmaxis": zmaxis,
        # "sibry": sibry,
        "fpol": fpol,
        "pres": pres,
        "ffprim": ffprim,
        "pprim": pprim,
        "psirz": psirz,
        "qpsi": qpsi,
        "bbsrz": bbsrz,
        "limrz": limrz,

    }


def sp_write_geqdsk(p, file):
    """
    :param profile: object

    :param file: file path / file
    :return:
    """
    if type(file) is str:
        file = open(file, "w")

    nw = p["nw"]
    nh = p["nh"]

    file.write("%48s%4i%4i%4i\n" % (p["description"], 3, p["nw"], p["nh"]))
    file.write("%16.9e%16.9e%16.9e%16.9e%16.9e\n" %
               (p["rdim"], p["zdim"], p["rcentr"], p["rleft"], p["zmid"]))
    file.write("%16.9e%16.9e%16.9e%16.9e%16.9e\n" %
               (p["rmaxis"], p["zmaxis"], p["simag"], p["sibry"], p["bcentr"]))
    file.write("%16.9e%16.9e%16.9e%16.9e%16.9e\n" %
               (p["current"], p["simag"], 0, p["rmaxis"], 0))
    file.write("%16.9e%16.9e%16.9e%16.9e%16.9e\n" %
               (p["zmaxis"], 0, p["sibry"], 0, 0))

    def _write_data(d):
        count = len(d)
        for n in range(count):
            file.write("%16.9e" % d[n])
            if (n == count - 1) or ((n + 1) % 5 == 0):
                file.write('\n')

    _write_data(p["fpol"])
    _write_data(p["pres"])
    _write_data(p["ffprim"])
    _write_data(p["pprim"])
    _write_data(p["psirz"].reshape([nw * nh]))
    _write_data(p["qpsi"])
    file.write("%5i%5i\n" % (p["bbsrz"].shape[0], p["limrz"].shape[0]))
    _write_data(p["bbsrz"].reshape([p["bbsrz"].size]))
    _write_data(p["limrz"].reshape([p["limrz"].size]))

    file.close()

    return


def sp_imas_equilibrium_to_geqdsk(eq, nw=125, nh=125):
    from scipy.interpolate import Meshdata

    coord_r = eq.coordinate_system.r
    coord_z = eq.coordinate_system.z
    rleft = coord_r.min()
    rdim = coord_r.max() - coord_r.min()
    zdim = coord_z.max() - coord_z.min()

    # rdim = 0.0
    # zdim = 0.0
    rcentr = eq.boundary.geometric_axis.r
    # rleft = 0.0
    zmid = eq.boundary.geometric_axis.z
    rmaxis = eq.global_quantities.magnetic_axis.r
    zmaxis = eq.global_quantities.magnetic_axis.z
    simag = eq.global_quantities.psi_axis
    sibry = eq.global_quantities.psi_boundary
    bcentr = eq.global_quantities.magnetic_axis.b_field_tor
    current = eq.global_quantities.ip

    # boundary

    rbbs = eq.boundary.lcfs.r
    zbbs = eq.boundary.lcfs.z

    bbsrz = np.append(rbbs.reshape([1, rbbs.size]), zbbs.reshape(
        [1, rbbs.size]), axis=0).transpose()
    # psi

    Mesh_r, Mesh_z = np.mMesh[rleft:rleft + rdim: nw *
                              1j, zmid - zdim / 2: zmid + zdim / 2: nh * 1j]
    coord_r = np.append(coord_r[:, :], coord_r[:, 0].reshape(
        coord_r.shape[0], 1), axis=1)
    coord_z = np.append(coord_z[:, :], coord_z[:, 0].reshape(
        coord_z.shape[0], 1), axis=1)
    points = np.append(coord_r.reshape(
        [coord_r.size, 1]), coord_z.reshape([coord_z.size, 1]), axis=1)
    psi = eq.profiles_2d[1].psi
    values = psi[:coord_r.shape[0], :coord_r.shape[1]].reshape(points.shape[0])
    psirz = Meshdata(points, values, (Mesh_r, Mesh_z),
                     method='cubic').transpose()

    # profile

    fpol = eq.profiles_1d.f
    pres = eq.profiles_1d.pressure
    ffprim = eq.profiles_1d.f_df_dpsi
    pprim = eq.profiles_1d.dpressure_dpsi
    qpsi = eq.profiles_1d.q
    print(qpsi.shape)

    return {
        "nw": nw,
        "nh": nh,
        "rdim": rdim,
        "zdim": zdim,
        "rcentr": rcentr,
        "rleft": rleft,
        "zmid": zmid,
        "rmaxis": rmaxis,
        "zmaxis": zmaxis,
        "simag": simag,
        "sibry": sibry,
        "bcentr": bcentr,
        "current": current,
        "bbsrz": bbsrz,
        "psirz": psirz,
        "fpol": fpol,
        "pres": pres,
        "ffprim": ffprim,
        "pprim": pprim,
        "qpsi": qpsi

    }


def sp_geqdsk_to_imas_equilibrium(geqdsk, eq):
    # rdim = 0.0
    # zdim = 0.0
    eq.boundary.geometric_axis.r = geqdsk["rcentr"]
    eq.boundary.geometric_axis.z = geqdsk["zmid"]
    # rleft = 0.0
    eq.global_quantities.magnetic_axis.r = geqdsk["rmaxis"]
    eq.global_quantities.magnetic_axis.z = geqdsk["zmaxis"]
    eq.global_quantities.psi_axis = geqdsk["simag"]
    eq.global_quantities.psi_boundary = geqdsk["sibry"]
    eq.global_quantities.magnetic_axis.b_field_tor = geqdsk["bcentr"]
    eq.global_quantities.ip = geqdsk["current"]

    # boundary

    eq.boundary.outline.r = geqdsk["bbsrz"][:, 0]
    eq.boundary.outline.z = geqdsk["bbsrz"][:, 1]

    eq.profiles_2d.resize(1)
    eq.profiles_2d[0].Mesh_type.name = "rectangular"
    eq.profiles_2d[0].Mesh_type.index = 1
    eq.profiles_2d[0].psi = geqdsk["psirz"]

    # coord_r = eq.coordinate_system.r
    # coord_z = eq.coordinate_system.z
    # rleft = coord_r.min()
    # rdim = coord_r.max() - coord_r.min()
    # zdim = coord_z.max() - coord_z.min()

    # bbsrz = np.append(rbbs.reshape([1, rbbs.size]), zbbs.reshape(
    #     [1, rbbs.size]), axis=0).transpose()
    # # psi

    # Mesh_r, Mesh_z = np.mMesh[rleft:rleft + rdim: nw *
    #                           1j, zmid - zdim / 2: zmid + zdim / 2: nh * 1j]
    # coord_r = np.append(coord_r[:, :], coord_r[:, 0].reshape(
    #     coord_r.shape[0], 1), axis=1)
    # coord_z = np.append(coord_z[:, :], coord_z[:, 0].reshape(
    #     coord_z.shape[0], 1), axis=1)
    # points = np.append(coord_r.reshape(
    #     [coord_r.size, 1]), coord_z.reshape([coord_z.size, 1]), axis=1)
    # psi = eq.profiles_2d[1].psi
    # values = psi[:coord_r.shape[0], :coord_r.shape[1]].reshape(points.shape[0])
    # psirz = Meshdata(points, values, (Mesh_r, Mesh_z),
    #                  method='cubic').transpose()

    # profile

    eq.profiles_1d.f = geqdsk["fpol"]
    eq.profiles_1d.pressure = geqdsk["pres"]
    eq.profiles_1d.f_df_dpsi = geqdsk["ffprim"]
    eq.profiles_1d.dpressure_dpsi = geqdsk["pprim"]
    eq.profiles_1d.q = geqdsk["qpsi"]


if __name__ == "__main__":
    import os
    os.environ['USER'] = 'fydev'

    pp = pprint.PrettyPrinter(indent=2)

    geqdsk = sp_read_geqdsk("/workspaces/example_data/g063982.04800")

    imas_obj = imas.ids(63982, 48)
    # Create a new instance of database
    imas_obj.create_env("fydev", "test", "3")
    imas_obj.equilibrium.ids_properties.homogeneous_time = 1
    imas_obj.equilibrium.Meshs_ggd.resize(1)
    imas_obj.equilibrium.Meshs_ggd[0].Mesh.resize(1)
    imas_obj.equilibrium.Meshs_ggd[0].Mesh[0].name = "unspecified"
    imas_obj.equilibrium.Meshs_ggd[0].Mesh[0].index = 0
    imas_obj.equilibrium.Meshs_ggd[0].time=1.234

    imas_obj.equilibrium.resize(1)
    imas_obj.equilibrium.time.resize(1)
    imas_obj.equilibrium.time[0] = 0.0

    sp_geqdsk_to_imas_equilibrium(geqdsk, imas_obj.equilibrium.time_slice[0])
    imas_obj.equilibrium.time_slice[0].time=0.0
    imas_obj.equilibrium.put()

    imas_obj.core_profiles.ids_properties.homogeneous_time = 1
    imas_obj.core_profiles.time.resize(1)
    imas_obj.core_profiles.time[0] = 0.0
    imas_obj.core_profiles.put()

    imas_obj.ec_launchers.ids_properties.homogeneous_time = 1
    imas_obj.ec_launchers.time.resize(1)
    imas_obj.ec_launchers.time[0] = 0.0
    imas_obj.ec_launchers.put()

    imas_obj.close()




import sys
sys.path.append("/home/salmon/workspace/freegs/")
sys.path.append("/home/salmon/workspace/fytok/python")
sys.path.append("/home/salmon/workspace/SpDev/SpDB")


if __name__ == '__main__':
    from ..util.logger import logger
    from spdm.core.Entry import open_entry

    entry = open_entry("east+mdsplus:///home/salmon/public_data/~t/?tree_name=efit_east", shot=55555, time_slice=10)

    # logger.debug(entry.pf_active.coil[1].element[0].geometry.rectangle.r())
    # logger.debug(entry.wall.description_2d[0].limiter.unit[0].outline.r())
    # logger.debug(entry.wall.description_2d[0].limiter.unit[0].outline.r())

    logger.debug(entry.equilibrium.vacuum_toroidal_field())
    logger.debug(entry.equilibrium.vacuum_toroidal_field.b0())
    # logger.debug(entry.equilibrium.profiles_1d())
    # logger.debug(entry.equilibrium.profiles_1d.dpressure_dpsi())



import collections.abc
import typing
from copy import copy

import numpy as np
from scipy.interpolate import (
    InterpolatedUnivariateSpline,
    RectBivariateSpline,
    RegularGridInterpolator,
    UnivariateSpline,
    interp1d,
    interp2d,
)

from ..core.Functor import Functor
from ..utils.logger import logger
from ..utils.typing import array_type


class RectInterpolateOp(Functor):
    def __init__(self, *xy, periods=None, check_nan=False, extrapolate=1, **kwargs) -> None:
        super().__init__(None)

        if len(xy) == 0:
            raise RuntimeError(f"Illegal dims={xy} ")

        self._value = xy[-1]
        self._dims = xy[:-1]
        self._periods = periods
        self._opts: dict = kwargs
        self._check_nan = check_nan
        self._extrapolate = extrapolate
        self._shape = tuple(len(d) for d in self._dims)
        self._ppoly = None
        if isinstance(self._value, array_type) and len(self._value.shape) > 0:
            if len(self._value.shape) > len(self._shape):
                raise NotImplementedError(
                    f"TODO: interpolate for rank >1 . { self._value.shape}!={self._shape}!  {xy} func={self.__str__()} "
                )
            elif tuple(self._value.shape) != tuple(self._shape):
                raise RuntimeError(
                    f"Function.compile() incorrect value shape { self._value.shape}!={self._shape}! func={self.__str__()} "
                )

    @property
    def ppoly(self):
        if self._ppoly is not None:
            return self._ppoly

        if len(self._dims) == 0:
            raise RuntimeError(f"Illegal dims={self._dims} ")

        value = self._value

        if callable(value):
            value = value(*np.meshgrid(*self._dims, indexing="ij"))

        if not isinstance(value, array_type) or not np.all(value.shape == self._shape):
            raise TypeError((value))

        elif len(self._dims) == 1:
            x = self._dims[0]

            if len(x) == 0:
                raise RuntimeError(f"x is empty!")

            if self._check_nan:
                mark = np.isnan(value)
                nan_count = np.count_nonzero(mark)

                if nan_count == len(value):
                    raise RuntimeError(f"value is NaN!")
                elif nan_count > 0:
                    # logger.warning(  f"{self.__class__.__name__}[{self.__str__()}]: Ignore {nan_count} NaN at {np.argwhere(mark)}.")
                    value = value[~mark]
                    x = x[~mark]

            try:
                self._ppoly = InterpolatedUnivariateSpline(x, value, ext=self._extrapolate)    
            except Exception as error:
                raise RuntimeError(f"Can not create Interpolator! \n x={x} value={value}") from error

        elif len(self._dims) == 2:
            if self._check_nan:
                mark = np.isnan(value)
                nan_count = np.count_nonzero(mark)
                if nan_count > 0:
                    logger.warning(
                        f"{self.__class__.__name__}[{self.__str__()}]: Replace  {nan_count} NaN by 0 at {np.argwhere(mark)}."
                    )
                    value[mark] = 0.0

            # if isinstance(self.periods, collections.abc.Sequence):
            #     logger.warning(f"TODO: periods={self.periods}")

            self._ppoly = RectBivariateSpline(*self._dims, value)
            self._opts = {"grid": False}

        else:
            self._ppoly = RegularGridInterpolator(self._dims, value)

        return self._ppoly

    def __call__(self, *args, **kwargs) -> typing.Any:
        try:
            res = self.ppoly(*args, **kwargs, **self._opts)
        except ValueError as e:
            raise RuntimeError(f"{self.__class__.__name__}[{self.__str__()}]") from e
        except TypeError as e:
            raise TypeError(f"{args}") from e

        if not isinstance(args[0], array_type) or args[0].size == 1:
            return np.squeeze(res, axis=0).item()
        else:
            return res

    def derivative(self, n=1) -> Functor:
        return Functor(self.ppoly.derivative(n), **self._opts)

    def partial_derivative(self, *d) -> Functor:
        if len(d) > 0:
            label = f"d_({self.__str__()})"
        else:
            label = f"d_{d}({self.__str__()})"

        return Functor(self.ppoly.partial_derivative(*d), label=label, **self._opts)

    def antiderivative(self, *d) -> Functor:
        return Functor(self.ppoly.antiderivative(*d), **self._opts)


def interpolate(*args, type="rectlinear", **kwargs) -> Functor:
    match type:
        case "rectlinear":
            res = RectInterpolateOp(*args, **kwargs)
        case _:
            raise NotImplementedError(f"type={type}")
    return res



import collections.abc
import typing
from io import BytesIO

import matplotlib.pyplot as plt
import numpy as np
from spdm.core.Path import update_tree, merge_tree
from spdm.core.Expression import Expression
from spdm.core.Signal import Signal
from spdm.core.Field import Field
from spdm.core.Function import Function
from spdm.geometry.BBox import BBox
from spdm.geometry.Circle import Circle
from spdm.geometry.Curve import Curve
from spdm.geometry.GeoObject import GeoObject
from spdm.geometry.Line import Line
from spdm.geometry.Point import Point
from spdm.geometry.PointSet import PointSet
from spdm.geometry.Polygon import Polygon, Rectangle
from spdm.geometry.Polyline import Polyline
from spdm.utils.envs import SP_DEBUG
from spdm.utils.logger import SP_DEBUG, logger
from spdm.utils.tags import _not_found_
from spdm.utils.typing import array_type, as_array, is_array, is_scalar
from spdm.view.View import View


@View.register(["matplotlib"])
class MatplotlibView(View):
    backend = "matplotlib"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _figure_post(
        self,
        fig: plt.Figure,
        title="",
        output=None,
        styles={},
        transparent=True,
        signature=None,
        width=1.0,
        height=1.0,
        **kwargs,
    ) -> typing.Any:
        fontsize = styles.get("fontsize", 16)

        fig.suptitle(title, fontsize=fontsize)

        fig.align_ylabels()

        fig.tight_layout()

        if signature is None:
            signature = self.signature

        if signature is not False:
            W, H = fig.get_size_inches()
            if H > 4:
                fig.text(
                    width,
                    0.0,  # 5 * height,
                    signature,
                    verticalalignment="bottom",
                    horizontalalignment="left",
                    fontsize="small",
                    alpha=0.2,
                    rotation="vertical",
                )
            else:
                fig.text(
                    width,
                    0.0,
                    signature,
                    verticalalignment="bottom",
                    horizontalalignment="right",
                    fontsize="small",
                    alpha=0.2,
                    # rotation="vertical",
                )

        if output == "svg":
            buf = BytesIO()
            fig.savefig(buf, format="svg", transparent=transparent, **kwargs)
            buf.seek(0)
            fig_html = buf.getvalue().decode("utf-8")
            plt.close(fig)
            fig = fig_html

        elif output is not None:
            logger.debug(f"Write figure to  {output}")
            kwargs.setdefault("format", "svg")
            fig.savefig(output, transparent=transparent, **kwargs)
            plt.close(fig)
            fig = None

        return fig

    def draw(self, geo, *styles, view_point="rz", title=None, scaled=False, **kwargs) -> typing.Any:
        fig, canvas = plt.subplots()

        geo = self._draw(canvas, geo, *styles, view_point=view_point)

        g_styles = geo.get("$styles", {}) if isinstance(geo, dict) else {}
        g_styles = merge_tree(g_styles, *styles)

        xlabel = g_styles.get("xlabel", None)

        if xlabel is not None:
            canvas.set_xlabel(xlabel)
        elif view_point.lower() == "rz":
            canvas.set_xlabel(r" $R$ [m]")
        else:
            canvas.set_xlabel(r" $X$ [m]")

        ylabel = g_styles.get("ylabel", None)
        if ylabel is not None:
            canvas.set_ylabel(ylabel)
        elif view_point.lower() == "rz":
            canvas.set_ylabel(r" $Z$ [m]")
        else:
            canvas.set_ylabel(r" $Y$ [m]")

        pos = canvas.get_position()

        canvas.set_aspect("equal")

        canvas.axis("scaled")

        new_pos = canvas.get_position()

        width = 1.0 + (new_pos.xmax - pos.xmax)
        height = 1.0 + (new_pos.ymax - pos.ymax)

        title = title or g_styles.get("title", None)

        return self._figure_post(fig, title=title, styles=g_styles, width=width, height=height, **kwargs)

    def _draw(
        self,
        canvas,
        geo: GeoObject | str | BBox,
        *styles,
        view_point=None,
        **kwargs,
    ):
        if False in styles:
            return
        g_styles = getattr(geo, "_metadata", {}).get("styles", {})
        g_styles = update_tree(g_styles, *styles)

        if geo is None or geo is _not_found_:
            pass
        elif hasattr(geo.__class__, "__geometry__"):
            try:
                geo = geo.__geometry__(view_point=view_point, **kwargs)
            except Exception as error:
                if SP_DEBUG == "strict":
                    raise RuntimeError(f"ignore unsupported geometry {geo.__class__.__name__} {geo}! ") from error
                else:
                    logger.exception(f"ignore unsupported geometry {geo.__class__.__name__} {geo}! ")

            else:
                self._draw(canvas, geo, *styles, view_point=view_point, **kwargs)

        elif hasattr(geo, "mesh") and hasattr(geo, "__array__"):
            self._draw(canvas, (*geo.mesh.points, geo.__array__()), *styles, **kwargs)

        elif isinstance(geo, dict):
            s_styles = geo.get("$styles", {})

            for s in [k for k in geo.keys() if not k.startswith("$")]:
                self._draw(canvas, geo[s], {"id": s}, s_styles, *styles, view_point=view_point, **kwargs)
            else:
                self._draw(canvas, geo.get("$data"), s_styles, *styles, view_point=view_point, **kwargs)

            self._draw(canvas, None, s_styles, *styles, **kwargs)

        elif isinstance(geo, list):
            for idx, g in enumerate(geo):
                self._draw(canvas, g, {"id": idx}, *styles, view_point=view_point, **kwargs)

            self._draw(canvas, None, *styles, view_point=view_point, **kwargs)

        elif geo.__class__ is GeoObject:
            self._draw(canvas, geo.bbox, *styles)

        else:
            s_styles = g_styles.get("$matplotlib", {})

            if isinstance(geo, tuple) and all([isinstance(g, array_type) for g in geo]):
                *x, y = geo

                if len(x) != 2 or y.ndim != 2:
                    raise RuntimeError(f"Illegal dimension {[d.shape for d in x]} {y.shape} ")

                canvas.contour(*x, y, **merge_tree(s_styles, {"levels": 20, "linewidths": 0.5}))

            elif isinstance(geo, (str, int, float, bool)):
                pos = g_styles.get("position", None)

                if pos is None:
                    return

                canvas.text(
                    *pos,
                    str(geo),
                    **collections.ChainMap(
                        s_styles,
                        {
                            "horizontalalignment": "center",
                            "verticalalignment": "center",
                            "fontsize": "xx-small",
                        },
                    ),
                )

            elif isinstance(geo, BBox):
                canvas.add_patch(plt.Rectangle(geo.origin, *geo.dimensions, fill=False, **s_styles))

            elif isinstance(geo, Polygon):
                canvas.add_patch(plt.Polygon(geo._points, fill=False, **s_styles))

            elif isinstance(geo, Polyline):
                canvas.add_patch(plt.Polygon(geo._points, fill=False, closed=geo.is_closed, **s_styles))

            elif isinstance(geo, Line):
                canvas.add_artist(plt.Line2D([geo.p0.x, geo.p1.x], [geo.p0.y, geo.p1.y], **s_styles))

            elif isinstance(geo, Curve):
                canvas.add_patch(plt.Polygon(geo._points, fill=False, closed=geo.is_closed, **s_styles))

            elif isinstance(geo, Rectangle):
                canvas.add_patch(plt.Rectangle((geo._x, geo._y), geo._width, geo._height, fill=False, **s_styles))

            elif isinstance(geo, Circle):
                canvas.add_patch(plt.Circle((geo.x, geo.y), geo.r, fill=False, **s_styles))

            elif isinstance(geo, Point):
                canvas.scatter(geo.x, geo.y, **s_styles)

            elif isinstance(geo, PointSet):
                canvas.scatter(*geo.points, **s_styles)

            else:
                raise RuntimeError(f"Unsupport type {(geo)} {geo}")

        text_styles = g_styles.get("text", False)

        if text_styles:
            if not isinstance(text_styles, dict):
                text_styles = {}

            if isinstance(geo, Line):
                text = geo.name
                pos = [geo.p1.x, geo.p1.y]
            elif isinstance(geo, GeoObject):
                text = geo.name
                pos = geo.bbox.center
            elif hasattr(geo, "mesh"):
                text = geo.name
                pos = geo.mesh.bbox.center
            else:
                text = str(geo)
                pos = None

            text_styles.setdefault("position", pos)

            self._draw(canvas, text, {f"${self.backend}": text_styles})

        return geo

    def plot(
        self,
        *args,
        x_axis: Expression | np.ndarray | str = None,
        x_label=None,
        styles=_not_found_,
        width=10,
        height=8,
        **kwargs,
    ) -> typing.Any:
        styles = update_tree({}, styles, kwargs)

        fontsize = styles.get("fontsize", 16)

        if len(args) > 1 and isinstance(args[0], array_type):
            x_value = args[0]
            args = args[1:]
        else:
            x_value = None

        nprofiles = len(args)

        height = max(2, height / nprofiles) * nprofiles

        fig, canvas = plt.subplots(ncols=1, nrows=nprofiles, sharex=True, figsize=(width, height))

        if nprofiles == 1:
            canvas = [canvas]

        if isinstance(x_axis, Expression):
            if x_label is None:
                units = x_axis._metadata.get("units", "-")
                x_label = f"{ x_value.__label__} [{units}]"
            if isinstance(x_value, array_type):
                x_axis = x_axis(x_value)
            elif x_value is None:
                x_value = as_array(x_axis)
                x_axis = None
        elif isinstance(x_axis, str):
            x_label = x_axis
            x_axis = None
        elif isinstance(x_axis, array_type):
            if x_value is None:
                x_value = x_axis
            elif x_value.size != x_value.size:
                raise RuntimeError(f"size mismatch {x_value.size} != {x_axis.size}")

        for idx, profiles in enumerate(args):
            if isinstance(profiles, tuple):
                profiles, sub_styles = profiles
            else:
                sub_styles = {}

            if sub_styles is False:
                continue
            elif isinstance(sub_styles, str):
                sub_styles = {"label": sub_styles}

            elif not isinstance(sub_styles, dict):
                raise RuntimeError(f"Unsupport sub_styles {sub_styles}")

            sub_styles = collections.ChainMap(sub_styles, styles)

            y_label = sub_styles.get("y_label", None)

            if not isinstance(profiles, (list)):
                profiles = [profiles]

            labels = []
            for p in profiles:
                if isinstance(p, tuple) and isinstance(p[1], (str, dict)):
                    p, p_styles = p
                else:
                    p_styles = {}

                if isinstance(p_styles, str) or p_styles is None:
                    p_styles = {"label": p_styles}

                p_styles = collections.ChainMap(p_styles, sub_styles)

                try:
                    t_label, t_y_label = self._plot(canvas[idx], x_value, p, x_axis=x_axis, styles=p_styles)

                    labels.append(t_label)
                    if y_label is None:
                        y_label = t_y_label

                except Exception as error:
                    if SP_DEBUG == "strict":
                        raise RuntimeError(f'Plot [index={idx}] failed! y_label= "{y_label}"  ') from error
                    else:
                        raise RuntimeError(f'Plot [index={idx}] failed! y_label= "{y_label}" ') from error

            if (vline := sub_styles.get("vline", _not_found_)) is not _not_found_:
                canvas[idx].axvline(**vline)

            if (hline := sub_styles.get("hline", _not_found_)) is not _not_found_:
                canvas[idx].axhline(**hline)

            if any(labels):
                canvas[idx].legend(fontsize=fontsize)
            if "$" not in y_label:
                y_label = f"${y_label}$"
            canvas[idx].set_ylabel(ylabel=y_label, fontsize=fontsize)

        if isinstance(x_label, str):
            if "$" not in x_label and "\\" in x_label:
                x_label = f"${x_label}$"

            canvas[-1].set_xlabel(x_label, fontsize=fontsize)

        return self._figure_post(fig, styles=styles, **kwargs)

    def _plot(self, canvas, x_value, expr, x_axis=None, styles=None, **kwargs) -> str:
        if expr is None or expr is _not_found_:
            return None, None

        styles = update_tree(kwargs, styles)

        s_styles = styles.get(f"${self.backend}", {})

        label = styles.get("label", None)

        y_value = None

        if isinstance(expr, Expression):
            if label is None:
                label = expr.__label__               
            y_value = expr(x_value)

        elif isinstance(expr, Signal):
            if x_value is None:
                y_value = expr.data
                x_value = expr.time
            else:
                y_value = expr(x_value)

            if label is None:
                label = expr.name

        elif isinstance(expr, array_type):
            y_value = expr

        elif hasattr(expr.__class__, "__array__"):
            y_value = expr.__array__()

        elif isinstance(expr, tuple) and len(expr) == 2 and all([isinstance(v, array_type) for v in expr]):
            x_value, y_value = expr

        else:
            y_value = expr

        if is_scalar(y_value):
            y_value = np.full_like(x_value, y_value, dtype=float)

        elif x_value is None:
            x_value = np.arange(len(expr))

        elif not isinstance(x_value, array_type):
            raise RuntimeError(f"ignore unsupported profiles label={label} {(y_value)}")

        if x_axis is None:
            x_axis = x_value

        if label is False:
            label = None
        elif not isinstance(label, str) or ("$" not in label and any(c in label for c in r"\{")):
            label = f"${label}$"

        canvas.plot(x_axis, y_value, **s_styles, label=label)

        units = getattr(expr, "_metadata", {}).get("units", "-")

        units = units.replace("^-1", "^{-1}").replace("^-2", "^{-2}").replace("^-3", "^{-3}").replace(".", " \cdot ")

        return label, f"[{units}]"

    # def profiles_(self, obj, *args,  x_axis=None, x=None,
    #               default_num_of_points=128, fontsize=10, grid=True,
    #               signature=None, title=None, **kwargs):
    #     fontsize = kwargs.get("fontsize", 10)
    #     nprofiles = len(obj)
    #     fig, canves = plt.subplots(
    #         ncols=1, nrows=nprofiles, sharex=True, figsize=(10, 2 * nprofiles)
    #     )
    #     self.draw(canves, obj, styles)
    #     x_label = kwargs.get("xlabel", "")
    #     if len(canves) == 1:
    #         canves[0].set_xlabel(x_label, fontsize=fontsize)
    #     else:
    #         canves[-1].set_xlabel(x_label, fontsize=fontsize)
    #     if not isinstance(profile_list, collections.abc.Sequence):
    #         profile_list = [profile_list]
    #     if isinstance(x_axis, collections.abc.Sequence) and not isinstance(
    #         x_axis, np.ndarray
    #     ):
    #         x_axis, x_label, *x_opts = x_axis
    #         x_opts = (x_opts or [{}])[0]
    #     else:
    #         x_axis = [0, 1]
    #         x_label = ""
    #         x_opts = {}
    #     if isinstance(x_axis, Function) and x is not None:
    #         x_axis = x_axis(x)
    #     elif x is None and isinstance(x_axis, np.ndarray):
    #         x = x_axis

    #     if isinstance(x_axis, np.ndarray):
    #         x_min = x_axis[0]
    #         x_max = x_axis[-1]
    #     elif isinstance(x_axis, collections.abc.Sequence) and len(x_axis) == 2:
    #         x_min, x_max = x_axis
    #         x_axis = np.linspace(x_min, x_max, default_num_of_points)
    #     else:
    #         raise TypeError(x_axis)

    #     if x is None and isinstance(x_axis, np.ndarray):
    #         x = x_axis
    #     elif callable(x_axis) or isinstance(x_axis, Function):
    #         x_axis = x_axis(x)

    #     nprofiles = len(profile_list)

    #     fig, sub_plot = plt.subplots(
    #         ncols=1, nrows=nprofiles, sharex=True, figsize=(10, 2 * nprofiles)
    #     )

    #     if not isinstance(sub_plot, (collections.abc.Sequence, np.ndarray)):
    #         sub_plot = [sub_plot]

    #     for idx, profile_grp in enumerate(profile_list):
    #         if not isinstance(profile_grp, list):
    #             profile_grp = [profile_grp]
    #         ylabel = None
    #         for jdx, p_desc in enumerate(profile_grp):
    #             profile, label, *o_args = p_desc
    #             opts = {}
    #             if len(o_args) > 0 and ylabel is None:
    #                 ylabel = o_args[0]
    #             if len(o_args) > 1:
    #                 opts = o_args[1]

    #             y = None

    #             if isinstance(profile, Function) or callable(profile):
    #                 try:
    #                     y = profile(x)
    #                 except Exception as error:
    #                     raise RuntimeError(
    #                         f"Can not get profile [idx={idx} jdx={jdx}]! name={getattr(profile,'_name',profile)}\n {error} "
    #                     ) from error

    #             elif isinstance(profile, np.ndarray) and len(profile) == len(x):
    #                 y = profile
    #             elif np.isscalar(profile):
    #                 y = np.full_like(x, profile, dtype=float)
    #             else:
    #                 raise RuntimeError(f"Illegal profile! {profile}!={x}")

    #             if not isinstance(y, np.ndarray) or not isinstance(x, np.ndarray):
    #                 logger.warning(f"Illegal profile! {(type(x) ,type(y), label, o_args)}")
    #                 continue
    #             elif x.shape != y.shape:
    #                 logger.warning(f"Illegal profile! {x.shape} !={y.shape}")
    #                 continue
    #             else:
    #                 # 删除 y 中的 nan
    #                 mark = np.isnan(y)
    #                 # if np.any(mark):
    #                 #     logger.warning(f"Found NaN in array  {np.argwhere(mark)}! {profile}  ")
    #                 sub_plot[idx].plot(x_axis[~mark], y[~mark], label=label, **opts)

    #         sub_plot[idx].legend(fontsize=fontsize)

    #         if grid:
    #             sub_plot[idx].grid()

    #         if ylabel is not None:
    #             sub_plot[idx].set_ylabel(ylabel, fontsize=fontsize)
    #         sub_plot[idx].labelsize = "media"
    #         sub_plot[idx].tick_params(labelsize=fontsize)

    #     if len(sub_plot) <= 1:
    #         sub_plot[0].set_xlabel(x_label, fontsize=fontsize)

    #     else:
    #         sub_plot[-1].set_xlabel(x_label, fontsize=fontsize)

    #     return fig

    # def draw_profile(self, profiles, x_axis, canves: plt.Axes = ..., style=None, **kwargs):
    #     if style is None:
    #         style = {}

    #     fontsize = style.get("fontsize", 10)

    #     ylabel = None

    #     x_value = x_axis

    #     if not isinstance(profiles, collections.abc.Sequence):
    #         profiles = [profiles]

    #     for profile, label, legend, *opts in profiles:
    #         y = None

    #         if isinstance(profile, Function) or callable(profile):
    #             try:
    #                 y = profile(x_value)
    #             except Exception as error:
    #                 raise RuntimeError(
    #                     f"Can not get profile! name={getattr(profile,'name',profile)}\n {error} "
    #                 ) from error

    #         elif isinstance(profile, array_type) and len(profile) == len(x_value):
    #             y = profile
    #         elif np.isscalar(profile):
    #             y = np.full_like(x_value, profile, dtype=float)
    #         else:
    #             raise RuntimeError(f"Illegal profile! {profile}!={x_value}")

    #         if not isinstance(y, array_type) or not isinstance(x_value, array_type):
    #             logger.warning(
    #                 f"Illegal profile! {(type(x_value) ,type(y), label, opts)}"
    #             )
    #             continue
    #         elif x.shape != y.shape:
    #             logger.warning(f"Illegal profile! {x_value.shape} !={y.shape}")
    #             continue
    #         else:
    #             # 删除 y 中的 nan
    #             mark = np.isnan(y)
    #             # if np.any(mark):
    #             #     logger.warning(f"Found NaN in array  {np.argwhere(mark)}! {profile}  ")
    #             canves.plot(x_axis[~mark], y[~mark], label=label, **opts)

    #     canves.legend(fontsize=fontsize)

    #     if kwargs.get("grid", True):
    #         canves.grid()

    #     if ylabel is not None:
    #         canves.set_ylabel(ylabel, fontsize=fontsize)
    #     canves.labelsize = "media"
    #     canves.tick_params(labelsize=fontsize)


# def sp_figure_signature(fig: plt.Figure, signature=None, x=1.0, y=0.1):
#     if signature is False:
#         return fig
#     elif not isinstance(signature, str):
#         signature = f"author: {getpass.getuser().capitalize()}. Create by SpDM at {datetime.datetime.now().isoformat()}."

#     pos = fig.gca().get_position()

#     fig.text(pos.xmax+0.01, 0.5*(pos.ymin+pos.ymax), signature,
#              verticalalignment='center', horizontalalignment='left',
#              fontsize='small', alpha=0.2, rotation='vertical')

#     # fig.text(x, y, signature, va='bottom', ha='left', fontsize='small', alpha=0.5, rotation='vertical')
#     return fig

# def plot(self, axis=None, *args, **kwargs):

#         if axis is None:
#             axis = plt.gca()

#         desc2d = self.description_2d[0]

#         # outline = desc2d.vessel.unit[0].annular.outline_inner

#         vessel_inner_points = np.array([desc2d.vessel.unit[0].annular.outline_inner.r,
#                                         desc2d.vessel.unit[0].annular.outline_inner.z]).transpose([1, 0])

#         vessel_outer_points = np.array([desc2d.vessel.unit[0].annular.outline_outer.r,
#                                         desc2d.vessel.unit[0].annular.outline_outer.z]).transpose([1, 0])

#         limiter_points = np.array([desc2d.limiter.unit[0].outline.r,
#                                    desc2d.limiter.unit[0].outline.z]).transpose([1, 0])

#         axis.add_patch(plt.Polygon(limiter_points, **
#                                    merge_tree(kwargs.get("limiter", {}), {"fill": False, "closed": True})))

#         axis.add_patch(plt.Polygon(vessel_outer_points, **merge_tree(kwargs.get("vessel_outer", {}),
#                                                                                kwargs.get("vessel", {}),
#                                                                                {"fill": False, "closed": True})))

#         axis.add_patch(plt.Polygon(vessel_inner_points, **merge_tree(kwargs.get("vessel_inner", {}),
#                                                                                kwargs.get("vessel", {}),
#                                                                                {"fill": False, "closed": True})))

#         return axis
#    def plot(self, axis=None, *args, with_circuit=False, **kwargs):

#         if axis is None:
#             axis = plt.gca()

#         for coil in self.coil:
#             rect = coil.element[0].geometry.rectangle

#             axis.add_patch(plt.Rectangle((rect.r - rect.width / 2.0,  rect.z - rect.height / 2.0),
#                                          rect.width,  rect.height,
#                                          **merge_tree(kwargs,  {"fill": False})))
#             axis.text(rect.r, rect.z, coil.name,
#                       horizontalalignment='center',
#                       verticalalignment='center',
#                       fontsize='xx-small')

#         return axis

# def plot(self, axis=None, *args, with_circuit=False, **kwargs):

#     if axis is None:
#         axis = plt.gca()
#     for idx, p_probe in enumerate(self.b_field_tor_probe):
#         pos = p_probe.position

#         axis.add_patch(plt.Circle((pos.r, pos.z), 0.01))
#         axis.text(pos.r, pos.z, idx,
#                   horizontalalignment='center',
#                   verticalalignment='center',
#                   fontsize='xx-small')

#     for p in self.flux_loop:
#         axis.add_patch(plt.Rectangle((p.position[0].r,  p.position[0].z), 0.01, 0.01))
#         axis.text(p.position[0].r, p.position[0].z, p.name,
#                   horizontalalignment='center',
#                   verticalalignment='center',
#                   fontsize='xx-small')
#     return axis



import unittest
from copy import deepcopy

from spdm.utils.tags import _not_found_
from spdm.view.View import display
from spdm.utils.logger import logger
from spdm.geometry.Circle import Circle


class TestView(unittest.TestCase):
    def test_display(self):
        c = Circle(0, 0, 1)
        logger.debug(display(c))


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, division, print_function

import numpy as np
from scipy import interpolate
from scipy.interpolate import CubicSpline, PPoly
from spdm.utils.logger import logger


def create_spline(x, y, **kwargs) -> PPoly:
    bc_type = "periodic" if np.all(y[0] == y[-1]) else "not-a-knot"
    try:
        res = CubicSpline(x, y, bc_type=bc_type)
    except ValueError as error:
        logger.error((x, y))
        raise error
    return res


def create_spline_for_bvp(y, yp, x, dx):
    """Create a cubic spline given values and derivatives.

    Formulas for the coefficients are taken from interpolate.CubicSpline.

    Returns
    -------
    sol : PPoly
        Constructed spline as a PPoly instance.
    """
    from scipy.interpolate import PPoly

    n, m = y.shape
    c = np.empty((4, n, m - 1), dtype=y.dtype)
    slope = (y[:, 1:] - y[:, :-1]) / dx
    t = (yp[:, :-1] + yp[:, 1:] - 2 * slope) / dx
    c[0] = t / dx
    c[1] = (slope - yp[:, :-1]) / dx - t
    c[2] = yp[:, :-1]
    c[3] = y[:, :-1]
    c = np.rollaxis(c, 1)

    return PPoly(c, x, extrapolate=True, axis=1)



import os

from spdm.utils.logger import logger
from spdm.core.Entry import open_entry

os.environ["SP_DATA_MAPPING_PATH"] = "/home/salmon/workspace/fytok_data/mapping"
DATA_PATH = "/home/salmon/workspace/fytok_data/"

if __name__ == '__main__':

    method = 1
    if method == 1:
        entry = open_entry(f"east+mdsplus://{DATA_PATH}/mdsplus/~t/?shot=70745")
    elif method == 2:
        entry = open_entry(f"EAST:///share/arch/east/~t/~f~e~d/?tree_name=east,t1,t2,t3,t4,t5,t6,shot=70754")
    else:
        os.environ['east_path'] = f'{DATA_PATH}/mdsplus/~t'
        os.environ['t1_path'] = f'{DATA_PATH}/mdsplus/~t'
        os.environ['t2_path'] = f'{DATA_PATH}/mdsplus/~t'
        os.environ['t3_path'] = f'{DATA_PATH}/mdsplus/~t'
        os.environ['t4_path'] = f'{DATA_PATH}/mdsplus/~t'
        os.environ['t5_path'] = f'{DATA_PATH}/mdsplus/~t'
        os.environ['t6_path'] = f'{DATA_PATH}/mdsplus/~t'

        entry = open_entry("MDSplus[EAST]:///#70754")

    pf_current = entry.child("pf_active/coil/0/current/data").find()

    pf_time = entry.child("pf_active/coil/0/current/time").find()

    # pprint({k: v for k, v in os.environ.items() if k.endswith("_path")})

    logger.debug(pf_current)

    logger.debug(pf_time)

    # time_slice = 100

    # eq = entry.child(f"equilibrium/time_slice/{time_slice}").fetch()

    # logger.debug(eq)

    logger.debug("DONE")



from spdm.core.sp_property import AttributeTree, SpTree, sp_property, sp_tree


class Data(SpTree):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class Foo(SpTree):
    boo: Data = sp_property(label="boo", default_value=1.0)


class Bar(Foo):
    boo: Data = sp_property(label="boo", units="m")


@sp_tree
class Bar2(Foo):
    boo: Data = 2.1234


a = Bar()
b = Bar2()
print(a.boo._metadata)
print(b.boo._metadata)
print(b.get("boo")._metadata)




from pprint import pprint


class Foo:
    def __setitem__(self, key, value):
        pprint(key)

    def __getitem__(self, key):
        pprint(key)


if __name__ == '__main__':

    foo = Foo()

    foo['a', 1:4] = 5

    foo['a', 1:4]



import os

from spdm.core.Entry import open_entry
from spdm.core.File import File

import matplotlib.pyplot as plt
from sptask import EQTASK


if __name__ == '__main__':
    # open entry for structure data and smi -structure data
    entry_db = open_entry("EAST://127.0.0.1#38300")

    # get pf,mag ,and wall
    wall = entry_db.child("wall")
    pf_active = entry_db.child("pf_active")
    magnetics = entry_db.child("magnetics")

    # open entry2 for non-structure data gdskfile ,which come from other simulation code.
    entry_file = open_entry("/<input path>/38300.gfile", format="GEqdsk", mode="r")

    time_slice = 50

    eq = entry_file.child("equilibrium/time_slice/{time_slice}")

    # Combine multiple IDSs into input_entry
    input_entry = {
        "wall": wall,
        "pf_active": pf_active,
        "magnetics": magnetics,
        "equilibrium": eq
    }

    # same entry to different eq code ,and unifie API GQTASK to call different eq code.

    # run the EQTASK to update out for eq_ids

    # Combine the new eq_ids and the unchanged device data into out_entry for the next step.

    task_1 = EQTASK(input_entry, id="freegs")

    result_1 = task_1.run()

    with File("/<output path>/38300_out.h5", mode="w") as fid:
        # open new entry to write IDS dict to hdf5 file
        fid.write(result_1)

    # plot the result directly

    task_2 = EQTASK(input_entry, id="atec")

    result_2 = task_2.run()

    plt.contour(result_2.get("time_slice/-1/profiles_2d/0/psi"))



from __future__ import annotations

import typing
from copy import copy

import numpy as np

from ..utils.logger import logger
from ..utils.typing import NumericType, as_array, numeric_type, scalar_type

ExprOpLike = typing.Callable | None


class Functor:
    """
    算符: 用于表示一个运算符，可以是函数，也可以是类的成员函数
    受 np.ufunc 启发而来。
    可以通过 ExprOp(op, method=method) 的方式构建一个 ExprOp 对象。
    """

    """
    ExprNode
    ---------
    ExprNode= Function + Node 是具有 Function 特性的 Node。
    Function 的 value 由 Node.__value__ 提供，

    mesh 可以根据 kwargs 中的 coordinate* 从 Node 所在 Tree 获得

    """

    def __init__(self, func: typing.Callable | None, /, method: str | None = None, label: str = None, **kwargs) -> None:
        self._func = func
        self._method = method
        self._label = label
        self._opts = kwargs

    def __copy__(self) -> Functor:
        """复制一个新的 Function 对象"""
        other: Functor = object.__new__(self.__class__)  # type:ignore
        other.__copy_from_(self)
        return other

    def __copy_from_(self, other: Functor) -> Functor:
        """复制一个新的 Function 对象"""
        self._func = other._func
        self._method = other._method
        self._label = other._label
        self._opts = copy(other._opts)
        return self

    @property
    def __label__(self) -> str:
        return self._label or str(self._func)

    @property
    def __annotation__(self) -> str:
        units = self._opts.get("units", "")
        label = self._opts.get("label", None)
        if label is not None:
            return f"{label}  {units}"
        else:
            return units

    def __domain__(self, *args, **kwargs):
        return True

    def __str__(self) -> str:
        return self.__label__

    def __call__(self, *args, **kwargs):
        if isinstance(getattr(self, "_method", None), str):
            op = getattr(self._func, self._method, None)
        elif self._method is not None:
            raise ValueError(self._method)
        else:
            op = self._func

        if not callable(op):
            return op

        try:
            kwargs.update(self._opts)
            value = op(*args, **kwargs)
        except Exception as error:
            raise RuntimeError(f"Error when apply  op={op} args={args} kwargs={kwargs}!") from error

        return value


class ConstantsFunc(Functor):
    def __init__(self, value: NumericType, **kwargs) -> None:
        super().__init__(None, **kwargs)
        if not isinstance(value, scalar_type):
            raise TypeError(f"value={value} is not a scalar!")
        self._value = value

    def __label__(self) -> str:
        return f"{self._value}"

    def __call__(self, *args, **kwargs):
        if len(args) > 0 and isinstance(args[0], np.ndarray):
            return np.full_like(args[0], self._value)
        else:
            return self._value


class SetpFun(Functor):
    def __init__(self, y: NumericType, *xargs, y0: NumericType = 0.0, **kwargs) -> None:
        super().__init__(None, **kwargs)
        self._y1 = y
        self._y0 = y0
        self._xargs = xargs

    def __label__(self) -> str:
        return r"H"

    def __call__(self, *args, **kwargs):
        return self._y1 if np.all(self._xargs > as_array(args)) else self._y0


class DiracDeltaFun(Functor):
    def __init__(self, y: NumericType, *xargs, y0: NumericType = 0.0, **kwargs) -> None:
        super().__init__(None, **kwargs)
        self._y1 = y
        self._y0 = y0
        self._xargs = xargs

    def __label__(self) -> str:
        return r"\delta"

    def derivative(self, n=1) -> SetpFun:
        if n == 1:
            return SetpFun(self._y1, self._xargs, y0=self._y0)
        else:
            raise NotImplementedError(f"n={n}")

    def __call__(self, *args, **kwargs):
        return self._y1 if np.allclose(self._xargs, as_array(args)) else self._y0


class DerivativeOp(Functor):
    """微分/积分算符"""

    def __init__(self, order, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._order = order

    def __copy__(self) -> DerivativeOp:
        res: DerivativeOp = super().__copy__()
        res._order = self._order
        return res

    @property
    def order(self) -> int | None:
        return self._order

        # return rf"\frac{{d({Expression._repr_s(self._children[1])})}}{{{Expression._repr_s(self._children[0])}}}"

    def _ppoly(self, *args, **kwargs):
        # if isinstance(self._expr, Variable):
        #     y = self._expr(*args)
        #     x = args[0]
        # elif isinstance(self._expr, Expression):
        #     x = self._expr.domain.points[0]
        #     y = self._expr(*args)
        return interpolate(*args, self._expr(*args)), args

    def __functor__(self):
        return self._expr.derivative(self._order)


def as_functor(expr, *args, **kwargs) -> Functor | None:
    if isinstance(expr, Functor):
        return expr
    elif callable(expr):
        return Functor(expr, *args, **kwargs)
    elif expr is None:
        return None
    else:
        raise TypeError(f"expr={expr} is not callable!")



from enum import Flag, auto


class tags(Flag):
    not_found = 0
    undefined = auto()
    empty = auto()
    next_ = auto()


_not_found_ = tags.not_found

_undefined_ = tags.undefined

_next_ = tags.next_

_empty = tags.empty



import typing
from .Actor import Actor
from dask.distributed import Client

_T = typing.TypeVar("_T")


class TaskManager:
    def __init__(self) -> None:
        self._client = Client()

    def submit_actor(self, cls: typing.Type[Actor]) -> None:

        return self._client.submit(cls, actor=True)



from spdm.data.NamedDict import NamedDict
from spdm.utils.logger import logger

if __name__ == '__main__':
    d = NamedDict({"a": 1, "b": {"c": 2}, "d": [1, 2, 3, 4, 5]})
    logger.debug(d.a)
    logger.debug(d.b.c)
    logger.debug(d.d)


from __future__ import annotations
import numpy as np
from .GeoObject import GeoObject
from .BBox import BBox
from ..utils.typing import array_type
from ..utils.logger import logger


class Box(GeoObject):
    """ Box 矩形，n维几何体
    """

    def __init__(self, xmin, xmax,  **kwargs) -> None:
        xmin = np.asarray(xmin)
        xmax = np.asarray(xmax)
        super().__init__(ndim=len(xmin),  **kwargs)
        self._xmin = xmin
        self._xmax = xmax
        self._bbox = BBox(xmin, xmax-xmin)

    @property
    def bbox(self) -> BBox: return self._bbox

    @property
    def is_closed(self) -> bool: return True
    @property
    def is_convex(self) -> bool: return True

    def enclose(self, *args) -> bool | array_type:
        """ Return True if all args are inside the geometry, False otherwise. """
        res = self.bbox.enclose(*args)
        if not np.sum(res):
            logger.debug((res, args, self._bbox))
        return res



import json
import typing

from spdm.core.Entry import Entry
from spdm.core.File import File
from spdm.utils.typing import as_native
from spdm.utils.logger import logger


@File.register(["json", "JSON"])
class FILEPLUGINjson(File):
    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._fid: typing.Optional[typing.IO[typing.Any]] = None

    def open(self) -> File:
        super().open()
        try:
            self._fid = open(self.path,  mode=self.mode_str)
        except OSError as error:
            raise FileExistsError(f"Can not open file {self.path}! {error}")
        else:
            logger.debug(
                f"Open {self.__class__.__name__} File {self.path} mode={self.mode}")
        return self

    def close(self):
        if self._fid is not None:
            self._fid.close()
            self._fid = None
        return super().close()

    def read(self, *args,   **kwargs) -> Entry:
        if not hasattr(self, "_fid"):
            self.open()
        return Entry(json.load(self._fid))

    def write(self,   d, *args,  **kwargs):
        json.dump(as_native(d, enable_ndarray=False), self._fid)


__SP_EXPORT__ = FILEPLUGINjson



from __future__ import annotations
from ..utils.plugin import Pluggable
from ..utils.logger import logger, SP_DEBUG

import collections
import collections.abc
import datetime
import getpass
import typing

import matplotlib.pyplot as plt
import os


#  在 MatplotlibView 中 imported matplotlib 会不起作用
#  报错 : AttributeError: module 'matplotlib' has no attribute 'colors'. Did you mean: 'colormaps'?


class View(Pluggable):
    """Abstract class for all views"""

    _plugin_registry = {}
    _plugin_prefix = "spdm.view.view_"

    backend = None

    # @classmethod
    # def __dispatch_init__(cls, _backend_type, self, *args, **kwargs) -> None:

    #     if isinstance(_backend_type, str):
    #         _backend_type = [_backend_type,
    #                          f"spdm.view.{_backend_type}#{_backend_type}",
    #                          f"spdm.view.{_backend_type}{cls.__name__}#{_backend_type}{cls.__name__}",
    #                          f"spdm.view.{_backend_type.capitalize()}#{_backend_type.capitalize()}",
    #                          f"spdm.view.{_backend_type.capitalize()}{cls.__name__}#{_backend_type.capitalize()}{cls.__name__}",
    #                          f"spdm.view.{cls.__name__}#{_backend_type}"
    #                          f"spdm.plugins.view.{_backend_type}#{_backend_type}",
    #                          f"spdm.plugins.view.{_backend_type}{cls.__name__}#{_backend_type}{cls.__name__}",
    #                          f"spdm.plugins.view.{_backend_type.capitalize()}#{_backend_type.capitalize()}",
    #                          f"spdm.plugins.view.{_backend_type.capitalize()}{cls.__name__}#{_backend_type.capitalize()}{cls.__name__}",
    #                          f"spdm.plugins.view.{cls.__name__}#{_backend_type}"
    #                          ]

    #     super().__dispatch_init__(_backend_type, self, *args, **kwargs)

    def __init__(self, *args, **kwargs) -> None:
        if self.__class__ is View:
            View.__dispatch_init__(kwargs.pop("backend", None), self, *args, **kwargs)
            return

    @property
    def signature(self) -> str:
        return f"Create by SpDM at {datetime.datetime.now().isoformat()}. AUTHOR: {getpass.getuser().capitalize()}. "

    def draw(self, *args, **kwargs):
        raise NotImplementedError(f"{self.__class__.__name__}.display")

    def plot(self, *args, **kwargs):
        raise NotImplementedError(f"{self.__class__.__name__}.draw")


_view_instances = {}


def viewer(backend=None):
    """Get a viewer instance"""

    if backend is None:
        backend = SP_VIEW_BACKEND  # "matplotlib"

    instance = _view_instances.get(backend, None)

    if instance is None:
        instance = _view_instances[backend] = View(backend=backend)

    return instance


SP_VIEW_BACKEND = "matplotlib"


def display(*args,   backend=None,  **kwargs):
    """Show an object"""

    return viewer(backend).draw(*args,  **kwargs)


def plot(*args,   backend=None, **kwargs):
    """Show an object"""
    return viewer(backend=backend).plot(*args,  **kwargs)



from importlib import resources as impresources
import collections
import collections.abc
import functools
import importlib
import importlib.util
import inspect
import os
import pathlib
import pkgutil
import sys
import typing

from .envs import SP_DEBUG
from .logger import logger

SP_EXPORT_KEYWORD = "__SP_EXPORT__"

SP_PATH: typing.List[pathlib.Path] = [pathlib.Path(p) for p in os.environ.get("SP_PATH", "").split(':')]


@functools.lru_cache
def sp_load_module(mod_name: str):

    module = sys.modules.get(mod_name, None)  # if module is loaded, use it

    if module is not None:
        return module

    spec = None

    try:
        spec = importlib.util.find_spec(mod_name)
    except ModuleNotFoundError:
        spec = None

    if spec is None:  # 在路径 SP_PATH 中搜索
        for p in SP_PATH:
            m_pth = p / mod_name.replace('.', '/')
            if m_pth.is_dir():
                m_pth = m_pth/"__init__.py"
            else:
                m_pth = m_pth.with_suffix(".py")

            if not (m_pth.exists() and m_pth.is_file()):
                continue

            try:
                spec = importlib.util.spec_from_file_location(mod_name, m_pth)
            except ModuleNotFoundError:
                spec = None
            else:
                if spec is not None:
                    break

    if spec is None:
        return None

    module = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(module)

    sys.modules[spec.name] = module

    if SP_DEBUG:
        logger.info(f"Load module {spec.name}")  # from {pathlib.Path(spec.origin).resolve().as_posix()}

    return module


def sp_find_module(path, fragment=None, pythonpath=None):
    if path is None:
        return None
    if not isinstance(path, str) and isinstance(path, collections.abc.Sequence):
        path = ".".join(path)

    o = path.replace('/', '.').strip('.').split('#')

    mod_name = o[0]
    if fragment is None and len(o) > 1:
        fragment = o[1]

    module = sys.modules.get(mod_name, None)  # if module is loaded, use it

    if module is None:
        if not pythonpath:
            try:
                spec = importlib.util.find_spec(mod_name)
            except ModuleNotFoundError:
                spec = None
        else:
            if isinstance(pythonpath, str):
                pythonpath = pythonpath.split(':')
            for p in pythonpath:
                mod_path = pathlib.Path(p) / mod_name.replace('.', '/')
                if mod_path.is_dir():
                    mod_path = mod_path/"__init__.py"
                else:
                    mod_path = mod_path.with_suffix(".py")

                if not (mod_path.exists() and mod_path.is_file()):
                    continue
                try:
                    spec = importlib.util.spec_from_file_location(mod_name, mod_path)
                except ModuleNotFoundError:
                    spec = None
                else:
                    if spec is not None:
                        break

        if spec is not None:
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            sys.modules[spec.name] = module

            if SP_DEBUG:
                logger.info(f"Load module {spec.name} from {spec.origin}")

    if not isinstance(module, object):
        module = None
    elif fragment is None:
        module = getattr(module, SP_EXPORT_KEYWORD, None) or getattr(
            module, path.split('.')[-1], None) or module
    elif hasattr(module, fragment):
        module = getattr(module, fragment)
    elif hasattr(module, SP_EXPORT_KEYWORD):
        module = getattr(getattr(module, SP_EXPORT_KEYWORD), fragment, None)
    else:
        module = None

    return module


def sp_find_subclass(cls, path: list):
    if len(path) == 0:
        return cls
    if not hasattr(cls, "__subclasses__"):
        return None

    for sub in cls.__subclasses__():
        if getattr(sub, "__name__", None) == path[0]:
            return sp_find_subclass(sub, path[1:])
    return None


def sp_pkg_data_path(pkg, rpath):
    if type(pkg) is str:
        pkg = sys.modules.get(pkg, None)

    plist = []
    if hasattr(pkg, "__path__"):  # check namespace package
        plist = pkg.__path__
    elif hasattr(pkg, "__file__"):  # check normal package
        plist = [pkg.__file__]
        yield pkgutil.get_data(pkg, rpath)
    # else:
    #     # raise ModuleNotFoundError(f"Module '{pkg}' is not loaded!")
    #     # return
    #     plist = []

    for p in plist:
        np = pathlib.Path(p)/rpath
        if np.exists():
            yield np


def export(fn):
    logger.debug((fn.__name__, fn.__package__))
    mod = sys.modules[fn.__module__]
    if hasattr(mod, '__all__'):
        mod.__all__.append(fn.__name__)
    else:
        mod.__all__ = [fn.__name__]
    return fn


def make_canonical_path_dot(path):
    if isinstance(path, str):
        path = path.replace("/", ".").strip(".")
    elif isinstance(path, collections.abc.Sequence):
        path = ".".join([p for p in path if p is not None])
    else:
        raise KeyError(f"Illegal path {path}")
    return path


def make_canonical_path_slash(path):
    if isinstance(path, str):
        path = path.replace(".", "/")
    elif isinstance(path, collections.abc.Sequence):
        path = "/".join([p for p in path if p is not None])
    else:
        raise KeyError(f"Illegal path {path}")
    return path


def make_canonical_path_list(path):
    if not isinstance(path, collections.abc.Sequence):
        path = [path]

    new_path = []
    for p in path:
        if isinstance(p, str):
            new_path.extend(p.split("/"))
        else:
            new_path.append(p)

    return [p for p in new_path if p is not None]


def absolute_path_dot(path, prefix):
    if isinstance(path, str):
        path = path.split('.')
    if isinstance(prefix, str):
        prefix = prefix.split('.')

    if path[0] == '':
        path = prefix.append(path)

    return ".".join([p for p in path if p != ''])


def absolute_path_slash(path, prefix):
    if isinstance(path, str):
        path = path.split('/')
    if isinstance(prefix, str):
        prefix = prefix.split('/')

    logger.debug((path, prefix))
    if path[0] != '':
        path = prefix.append(path)

    logger.debug(path)

    return "/"+"/".join([p for p in path if p != ''])


def relative_module_path(cls, base):
    return [p.__name__.lower()
            for p in inspect.getmro(cls) if issubclass(p, base) and p is not base and p is not cls][:: -1]+[cls.__name__]


def walk_namespace_modules(namespace: str, ignores=["__pycache__", "obsolete"]):
    """
        Walks a namespace and returns a list of all modules found within it.
        This function is useful for discovering all modules within a package.
        :param namespace: The namespace to walk.        
        :param ignores: A list of directories to ignore.
    """

    namespace = namespace.strip('.')

    try:
        contents = impresources.contents(namespace)
    except Exception:
        return

    if "__init__.py" in contents:
        yield namespace
    else:
        for sub in impresources.contents(namespace):
            if sub in ignores:
                continue
            elif sub.endswith(".py"):
                yield namespace + "." + sub[:-3]
            else:
                yield from walk_namespace_modules(namespace + "." + sub, ignores=ignores)



import unittest
from copy import deepcopy
from logging import log

from spdm.utils.logger import logger
from spdm.utils.tags import _not_found_
from spdm.core.Entry import Entry
from spdm.core.Mapper import MapperPath


class TestEntryMap(unittest.TestCase):
    data = {
        "a": [
            "hello world {name}!",
            "hello world2 {name}!",
            1, 2, 3, 4
        ],
        "c": "I'm {age}!",
        "d": {
            "e": "{name} is {age}",
            "f": "{address}"
        }
    }
    mapping = """<?xml version="1.0" ?>
    <root>
        <first> a/0 </first>
        <second> a/1 </second>
        <others> c/2:5 </others>
        <words>
            <address>
               [c,d/e,d/f]
            </address>
        </words>
    </root>
    """

    def test_mapping_path(self):
        pathmapper = MapperPath(Entry(self.mapping, scheme="XML"))

        self.assertEqual((pathmapper/"first").as_request(), "a/0")
        self.assertEqual((pathmapper/"second").as_request(), "a/1")
        self.assertEqual((pathmapper/"others").as_request(), "c/2:5")
        self.assertEqual((pathmapper/"words/address").as_request(), "[c,d/e,d/f]")

    def test_get(self):
        entry = EntryMapper(self.data, PathMapper(Entry(self.mapping, scheme="XML")))
        self.assertEqual(entry.get("first"), self.data["a"][0])
        self.assertEqual(entry.get("second"), self.data["a"][1])
        self.assertEqual(entry.get("others"), self.data["c"][2:5])


if __name__ == '__main__':
    unittest.main()
